# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Log to `Weights and Biases <https://wandb.ai/>`_."""

from __future__ import annotations

import atexit
import copy
import os
import pathlib
import re
import sys
import tempfile
import textwrap
import warnings
from typing import TYPE_CHECKING, Any, Optional, Sequence, Union

import numpy as np
import torch

from composer.loggers.logger import Logger
from composer.loggers.logger_destination import LoggerDestination
from composer.utils import MissingConditionalImportError, dist

if TYPE_CHECKING:
    from composer.core import State

__all__ = ['WandBLogger']


class WandBLogger(LoggerDestination):
    """Log to `Weights and Biases <https://wandb.ai/>`_.

    Args:
        project (str, optional): WandB project name.
        group (str, optional): WandB group name.
        name (str, optional): WandB run name.
            If not specified, the :attr:`.State.run_name` will be used.
        entity (str, optional): WandB entity name.
        tags (list[str], optional): WandB tags.
        log_artifacts (bool, optional): Whether to log
            `artifacts <https://docs.wandb.ai/ref/python/artifact>`_ (Default: ``False``).
        rank_zero_only (bool, optional): Whether to log only on the rank-zero process.
            When logging `artifacts <https://docs.wandb.ai/ref/python/artifact>`_, it is
            highly recommended to log on all ranks.  Artifacts from ranks ≥1 will not be
            stored, which may discard pertinent information. For example, when using
            Deepspeed ZeRO, it would be impossible to restore from checkpoints without
            artifacts from all ranks (default: ``True``).
        init_kwargs (dict[str, Any], optional): Any additional init kwargs
            ``wandb.init`` (see
            `WandB documentation <https://docs.wandb.ai/ref/python/init>`_).
    """

    def __init__(
        self,
        project: Optional[str] = None,
        group: Optional[str] = None,
        name: Optional[str] = None,
        entity: Optional[str] = None,
        tags: Optional[list[str]] = None,
        log_artifacts: bool = False,
        rank_zero_only: bool = True,
        init_kwargs: Optional[dict[str, Any]] = None,
    ) -> None:
        try:
            import wandb
        except ImportError as e:
            raise MissingConditionalImportError(
                extra_deps_group='wandb',
                conda_package='wandb',
                conda_channel='conda-forge',
            ) from e

        del wandb  # unused
        if log_artifacts and rank_zero_only and dist.get_world_size() > 1:
            warnings.warn((
                'When logging artifacts, `rank_zero_only` should be set to False. '
                'Artifacts from other ranks will not be collected, leading to a loss of information required to '
                'restore from checkpoints.'
            ))
        self._enabled = (not rank_zero_only) or dist.get_global_rank() == 0

        if init_kwargs is None:
            init_kwargs = {}

        if project is not None:
            init_kwargs['project'] = project

        if group is not None:
            init_kwargs['group'] = group

        if name is not None:
            init_kwargs['name'] = name

        if entity is not None:
            init_kwargs['entity'] = entity

        if tags is not None:
            init_kwargs['tags'] = tags

        self._rank_zero_only = rank_zero_only
        self._log_artifacts = log_artifacts
        self._init_kwargs = init_kwargs
        self._is_in_atexit = False

        # Set these variable directly to allow fetching an Artifact **without** initializing a WandB run
        # When used as a LoggerDestination, these values are overriden from global rank 0 to all ranks on Event.INIT
        self.entity = entity
        self.project = project

        self.run_dir: Optional[str] = None
        self.run_url: Optional[str] = None

        self.table_dict = {}

    def _set_is_in_atexit(self):
        self._is_in_atexit = True

    def log_hyperparameters(self, hyperparameters: dict[str, Any]):
        if self._enabled:
            import wandb
            wandb.config.update(hyperparameters)

    def log_table(
        self,
        columns: list[str],
        rows: list[list[Any]],
        name: str = 'Table',
        step: Optional[int] = None,
    ) -> None:
        if self._enabled:
            import wandb
            table = wandb.Table(columns=columns, rows=rows)
            wandb.log({name: table}, step=step)

    def log_metrics(self, metrics: dict[str, Any], step: Optional[int] = None) -> None:
        if self._enabled:
            import wandb

            # wandb.log alters the metrics dictionary object, so we deepcopy to avoid
            # side effects.
            metrics_copy = copy.deepcopy(metrics)
            wandb.log(metrics_copy, step)

    def log_images(
        self,
        images: Union[np.ndarray, torch.Tensor, Sequence[Union[np.ndarray, torch.Tensor]]],
        name: str = 'Images',
        channels_last: bool = False,
        step: Optional[int] = None,
        masks: Optional[dict[str, Union[np.ndarray, torch.Tensor, Sequence[Union[np.ndarray, torch.Tensor]]]]] = None,
        mask_class_labels: Optional[dict[int, str]] = None,
        use_table: bool = False,
    ):
        if self._enabled:
            import wandb
            if not isinstance(images, Sequence) and images.ndim <= 3:
                images = [images]

            # _convert_to_wandb_image doesn't include wrapping with wandb.Image to future
            # proof for when we support masks.
            images_generator = (_convert_to_wandb_image(image, channels_last) for image in images)

            if masks is not None:
                # Create a generator that yields masks in the format wandb wants.
                wandb_masks_generator = _create_wandb_masks_generator(
                    masks,
                    mask_class_labels,
                    channels_last=channels_last,
                )
                wandb_images = (
                    wandb.Image(im, masks=mask_dict) for im, mask_dict in zip(images_generator, wandb_masks_generator)
                )

            else:
                wandb_images = (wandb.Image(image) for image in images_generator)

            if use_table:
                table = wandb.Table(columns=[name])
                for wandb_image in wandb_images:
                    table.add_data(wandb_image)
                wandb.log({name + ' Table': table}, step=step)
            else:
                wandb.log({name: list(wandb_images)}, step=step)

    def init(self, state: State, logger: Logger) -> None:
        import wandb
        del logger  # unused

        # Use the state run name if the name is not set.
        if 'name' not in self._init_kwargs or self._init_kwargs['name'] is None:
            self._init_kwargs['name'] = state.run_name

        # Adjust name and group based on `rank_zero_only`.
        if not self._rank_zero_only:
            name = self._init_kwargs['name']
            self._init_kwargs['name'] += f'-rank{dist.get_global_rank()}'
            self._init_kwargs['group'] = self._init_kwargs['group'] if 'group' in self._init_kwargs else name
        if self._enabled:
            wandb.init(**self._init_kwargs)
            assert wandb.run is not None, 'The wandb run is set after init'
            entity_and_project = [str(wandb.run.entity), str(wandb.run.project)]
            self.run_dir = wandb.run.dir
            self.run_url = wandb.run.get_url()
            atexit.register(self._set_is_in_atexit)
        else:
            entity_and_project = [None, None]
        # Share the entity and project across all ranks, so they are available on ranks that did not initialize wandb
        dist.broadcast_object_list(entity_and_project)
        self.entity, self.project = entity_and_project
        assert self.entity is not None, 'entity should be defined'
        assert self.project is not None, 'project should be defined'

    def upload_file(self, state: State, remote_file_name: str, file_path: pathlib.Path, *, overwrite: bool):
        del overwrite  # unused

        if self._enabled and self._log_artifacts:
            import wandb

            # Some WandB-specific alias extraction
            timestamp = state.timestamp
            aliases = ['latest', f'ep{int(timestamp.epoch)}-ba{int(timestamp.batch)}']

            # replace all unsupported characters with periods
            # Only alpha-numeric, periods, hyphens, and underscores are supported by wandb.
            new_remote_file_name = re.sub(r'[^a-zA-Z0-9-_\.]', '.', remote_file_name)
            if new_remote_file_name != remote_file_name:
                warnings.warn((
                    'WandB permits only alpha-numeric, periods, hyphens, and underscores in file names. '
                    f"The file with name '{remote_file_name}' will be stored as '{new_remote_file_name}'."
                ))

            extension = new_remote_file_name.split('.')[-1]

            metadata = {f'timestamp/{k}': v for (k, v) in state.timestamp.state_dict().items()}
            # if evaluating, also log the evaluation timestamp
            if state.dataloader is not state.train_dataloader:
                # TODO If not actively training, then it is impossible to tell from the state whether
                # the trainer is evaluating or predicting. Assuming evaluation in this case.
                metadata.update({f'eval_timestamp/{k}': v for (k, v) in state.eval_timestamp.state_dict().items()})

            # Change the extension so the checkpoint is compatible with W&B's model registry
            if extension == 'pt':
                extension = 'model'

            wandb_artifact = wandb.Artifact(
                name=new_remote_file_name,
                type=extension,
                metadata=metadata,
            )
            wandb_artifact.add_file(os.path.abspath(file_path))
            wandb.log_artifact(wandb_artifact, aliases=aliases)

    def can_upload_files(self) -> bool:
        """Whether the logger supports uploading files."""
        return True

    def download_file(
        self,
        remote_file_name: str,
        destination: str,
        overwrite: bool = False,
        progress_bar: bool = True,
    ):
        # Note: WandB doesn't support progress bars for downloading
        del progress_bar
        import wandb
        import wandb.errors

        # using the wandb.Api() to support retrieving artifacts on ranks where
        # artifacts are not initialized
        api = wandb.Api()
        if not self.entity or not self.project:
            raise RuntimeError('get_file_artifact can only be called after running init()')

        # replace all unsupported characters with periods
        # Only alpha-numeric, periods, hyphens, and underscores are supported by wandb.
        if ':' not in remote_file_name:
            remote_file_name += ':latest'

        new_remote_file_name = re.sub(r'[^a-zA-Z0-9-_\.:]', '.', remote_file_name)
        if new_remote_file_name != remote_file_name:
            warnings.warn((
                'WandB permits only alpha-numeric, periods, hyphens, and underscores in file names. '
                f"The file with name '{remote_file_name}' will be stored as '{new_remote_file_name}'."
            ))

        try:
            wandb_artifact = api.artifact('/'.join([self.entity, self.project, new_remote_file_name]))
        except wandb.errors.CommError as e:
            raise FileNotFoundError(f'WandB Artifact {new_remote_file_name} not found') from e
        with tempfile.TemporaryDirectory() as tmpdir:
            wandb_artifact_folder = os.path.join(tmpdir, 'wandb_artifact_folder/')
            wandb_artifact.download(root=wandb_artifact_folder)
            wandb_artifact_names = os.listdir(wandb_artifact_folder)
            # We only log one file per artifact
            if len(wandb_artifact_names) > 1:
                raise RuntimeError(
                    'Found more than one file in WandB artifact. We assume the checkpoint is the only file in the WandB artifact.',
                )
            wandb_artifact_name = wandb_artifact_names[0]
            wandb_artifact_path = os.path.join(wandb_artifact_folder, wandb_artifact_name)
            if overwrite:
                os.replace(wandb_artifact_path, destination)
            else:
                os.rename(wandb_artifact_path, destination)

    def post_close(self) -> None:
        import wandb

        # Cleaning up on post_close so all artifacts are uploaded
        if not self._enabled or wandb.run is None or self._is_in_atexit:
            # Don't call wandb.finish if there is no run, or
            # the script is in an atexit, since wandb also hooks into atexit
            # and it will error if wandb.finish is called from the Composer atexit hook
            # after it is called from the wandb atexit hook
            return

        exc_tpe, exc_info, tb = sys.exc_info()

        if (exc_tpe, exc_info, tb) == (None, None, None):
            wandb.finish(0)
        else:
            # record there was an error
            wandb.finish(1)


def _convert_to_wandb_image(image: Union[np.ndarray, torch.Tensor], channels_last: bool) -> np.ndarray:
    if isinstance(image, torch.Tensor):
        if image.dtype == torch.float16 or image.dtype == torch.bfloat16:
            image = image.data.cpu().to(torch.float32).numpy()
        else:
            image = image.data.cpu().numpy()

    # Error out for empty arrays or weird arrays of dimension 0.
    if np.any(np.equal(image.shape, 0)):
        raise ValueError(f'Got an image (shape {image.shape}) with at least one dimension being 0! ')

    # Squeeze any singleton dimensions and then add them back in if image dimension
    # less than 3.
    image = image.squeeze()

    # Add in length-one dimensions to get back up to 3
    # putting channels last.
    if image.ndim == 1:
        image = np.expand_dims(image, (1, 2))
        channels_last = True
    if image.ndim == 2:
        image = np.expand_dims(image, 2)
        channels_last = True

    if image.ndim != 3:
        raise ValueError(
            textwrap.dedent(
                f'''Input image must be 3 dimensions, but instead
                            got {image.ndim} dims at shape: {image.shape}
                            Your input image was interpreted as a batch of {image.ndim}
                            -dimensional images because you either specified a
                            {image.ndim + 1}D image or a list of {image.ndim}D images.
                            Please specify either a 4D image of a list of 3D images''',
            ),
        )
    assert isinstance(image, np.ndarray)
    if not channels_last:
        image = image.transpose(1, 2, 0)
    return image


def _convert_to_wandb_mask(mask: Union[np.ndarray, torch.Tensor], channels_last: bool) -> np.ndarray:
    mask = _convert_to_wandb_image(mask, channels_last)
    mask = mask.squeeze()
    if mask.ndim != 2:
        raise ValueError(f'Mask must be a 2D array, but instead got array of shape: {mask.shape}')
    return mask


def _preprocess_mask_data(
    masks: dict[str, Union[np.ndarray, torch.Tensor, Sequence[Union[np.ndarray, torch.Tensor]]]],
    channels_last: bool,
) -> dict[str, np.ndarray]:
    preprocesssed_masks = {}
    for mask_name, mask_data in masks.items():
        if not isinstance(mask_data, Sequence):
            mask_data = mask_data.squeeze()
            if mask_data.ndim == 2:
                mask_data = [mask_data]
        preprocesssed_masks[mask_name] = np.stack([_convert_to_wandb_mask(mask, channels_last) for mask in mask_data])
    return preprocesssed_masks


def _create_wandb_masks_generator(
    masks: dict[str, Union[np.ndarray, torch.Tensor, Sequence[Union[np.ndarray, torch.Tensor]]]],
    mask_class_labels: Optional[dict[int, str]],
    channels_last: bool,
):
    preprocessed_masks: dict[str, np.ndarray] = _preprocess_mask_data(masks, channels_last)
    for all_masks_for_single_example in zip(*list(preprocessed_masks.values())):
        mask_dict = {name: {'mask_data': mask} for name, mask in zip(masks.keys(), all_masks_for_single_example)}
        if mask_class_labels is not None:
            for k in mask_dict.keys():
                mask_dict[k].update({'class_labels': mask_class_labels})
        yield mask_dict



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""FX-based model transformation and optimization.

Provides utilities to do FX-based model transformations.
"""

import logging
import operator
import re
from typing import Any, Callable, Mapping, Optional, Union

import torch
import torch.nn as nn
from torch.fx import GraphModule, Node
from torch.fx.passes.split_utils import split_by_tags

from composer.algorithms.stochastic_depth.stochastic_layers import BlockStochasticModule
from composer.utils import ensure_tuple

log = logging.getLogger(__name__)

__all__ = ['count_op_instances', 'replace_op', 'fuse_parallel_linears', 'apply_stochastic_residual']


def count_op_instances(gm: GraphModule, ops: Union[Callable, str, list[Union[Callable, str]]]) -> int:
    """Counts the number of instances of ``op`` in ``gm``.

    .. rubric:: Example

    .. testsetup::

        import operator
        import torch
        from torch.fx import symbolic_trace
        from composer.utils.fx_utils import count_op_instances

    .. doctest::

        >>> class M(torch.nn.Module):
        ...   def forward(self, x, y):
        ...     return x + y, torch.add(x, y), x.add(y)
        >>> module = M()
        >>> traced = symbolic_trace(module)
        >>> count_op_instances(traced, torch.add)
        1
        >>> count_op_instances(traced, [operator.add, torch.add, "add"])
        3

    Arguments:
        module (GraphModule): The source FX-traced graph.
        op (Union[Callable, str, list[Union[Callable, str]]]):
            The operations to count.

    Returns:
        int: The number of instances of ``ops`` in ``gm``
    """
    ops = list(ensure_tuple(ops))
    all_modules = dict(gm.named_modules())
    count = 0
    for n in gm.graph.nodes:
        for op in ops:
            if n.target == op:
                count += 1
            elif n.op == 'call_module' and isinstance(op, type) and isinstance(n.target, str) and isinstance(
                all_modules[n.target],
                op,
            ):
                count += 1
    return count


def replace_op(
    gm: GraphModule,
    src_ops: Union[Callable, str, list[Union[Callable, str]]],
    tgt_op: Callable[..., Any],
) -> GraphModule:
    """Replace a single operator, torch method or function with another.

    .. rubric:: Example

    .. testsetup::

        import operator
        import torch
        from torch.fx import symbolic_trace
        from composer.utils.fx_utils import replace_op, count_op_instances

    .. doctest::

        >>> class M(torch.nn.Module):
        ...   def forward(self, x, y):
        ...     return x + y, torch.add(x, y), x.add(y)
        >>> module = M()
        >>> traced = symbolic_trace(module)
        >>> traced = replace_op(traced, [operator.add, torch.add, "add"], torch.mul)
        >>> count_op_instances(traced, torch.mul)
        3

    Arguments:
        module (GraphModule): The source FX-traced graph.
        src_ops (Union[Callable, str, list[Union[Callable, str]]):
            Replace these operations.
        tgt_op (Callable): Replacement for the operations

    Returns:
        GraphModule: Modified GraphModule with each instance of an op in ``src_ops`` replaced with
            ``tgt_op``. Returns the input if no instances are found.
    """
    src_ops = list(ensure_tuple(src_ops))
    for n in gm.graph.nodes:
        if any(n.target == op for op in src_ops):
            with gm.graph.inserting_after(n):
                new_node = gm.graph.call_function(tgt_op, n.args, n.kwargs)
                n.replace_all_uses_with(new_node)
            gm.graph.erase_node(n)
    gm.recompile()
    return gm


def _get_ancestors(node: Node) -> list[Node]:
    ancestorNodes = []
    while node.op != 'placeholder':
        ancestorNodes.append(node)
        node = node.all_input_nodes[0]
    return ancestorNodes


def _get_residual_block_nodes(nodeLHS: Node, nodeRHS: Node) -> tuple[list[Node], list[Node]]:
    """Walk backwards from nodeLHS and nodeRSH to the root and construct lists of their parents.

    Arguments:
        nodeLHS (Node): left-hand side node for a binary operator
        nodeRHS (Node): right-hand side node for a binary operator

    Returns:
        (lhsAncestors, rhsAncestors): Two lists of nodes containing ancestors for ``nodeLHS`` and ``nodeRHS`` with
            their common ancestors removed.
    """
    lhsAncestors = _get_ancestors(nodeLHS)
    rhsAncestors = _get_ancestors(nodeRHS)

    # Iterate from back and eliminate common nodes
    while lhsAncestors and rhsAncestors and lhsAncestors[-1] == rhsAncestors[-1]:
        lhsAncestors.pop()
        rhsAncestors.pop()
    lhsAncestors.reverse()
    rhsAncestors.reverse()
    return lhsAncestors, rhsAncestors


def _attach_tag(nodes: list[Node], tag: str):
    """Attach tag to the given nodes for the splitter."""
    for node in nodes:
        node.tag = tag  # type: ignore[attr-defined]


def _tag_residual_nodes(gm: GraphModule) -> tuple[list[str], int]:
    """Tag nodes for splitting."""
    # all nodes that are not a part of the residual blocks are tagged with "mainN_{count}".
    # a tag is required for all nodes by split_by_tags
    # Also an earlier tag can be repeated for later nodes.
    count = 0
    all_tags = []
    # In this pass over all nodes, we just tag them
    for node in gm.graph.nodes:
        default_tag = f'mainN_{count}'
        node.tag = default_tag
        if default_tag not in all_tags:
            all_tags.append(default_tag)
        if node.op == 'call_function' and node.target in [torch.add, operator.add]:
            assert len(node.all_input_nodes) == 2
            node0, node1 = node.all_input_nodes[0], node.all_input_nodes[1]
            lhs_nodes, rhs_nodes = _get_residual_block_nodes(node0, node1)
            if lhs_nodes or rhs_nodes:
                if len(lhs_nodes):
                    _attach_tag(lhs_nodes, f'non_res_{count}')
                    all_tags.append(f'non_res_{count}')
                if len(rhs_nodes):
                    _attach_tag(rhs_nodes, f'residual_{count}')
                    all_tags.append(f'residual_{count}')
                add_tag = f'addN_{count}'
                if add_tag not in all_tags:
                    all_tags.append(add_tag)
                node.tag = add_tag
                count += 1
    return all_tags, count


def _get_residual_modules(gm: GraphModule, node: Node) -> tuple[Optional[GraphModule], Optional[GraphModule], int]:
    """Returns GraphModules for the main and residual branches.

    node.op is assumed to be a call_module
    """
    pattern = re.compile(r'non_res_(\d+)|residual_(\d+)')
    matches = pattern.match(str(node.target))
    if matches:
        idx = int(matches[1]) if matches[1] else int(matches[2])
        main_submod = getattr(gm, f'non_res_{idx}')
        residual_submod = getattr(gm, f'residual_{idx}', None)
        return main_submod, residual_submod, idx
    else:
        return None, None, 0


def _replace_residual_pattern(
    gm: GraphModule,
    original_node: Node,
    replacement_module: str,
    has_residual_ops: bool = False,
) -> None:
    """Replaces main, residual and add_node with the ``replacement_module``.

    ``replacement_module`` is already added to the gm.
    """
    insert_node = original_node.prev
    add_node = original_node.next
    if has_residual_ops:
        add_node = original_node.next.next
    with gm.graph.inserting_after(insert_node):
        new_node = gm.graph.call_module(replacement_module, args=(insert_node,))  # type: ignore
        add_node.replace_all_uses_with(new_node)
        gm.graph.erase_node(add_node)
        if has_residual_ops:
            gm.graph.erase_node(original_node.next)
        gm.graph.erase_node(original_node)
    gm.graph.lint()


def apply_stochastic_residual(gm: GraphModule, drop_rate: float = 0.2) -> tuple[GraphModule, int]:
    """Detect and replace residual pattern with their stochastic equivalent.

    Arguments:
        gm (GraphModule): The source FX-traced graph. It can be the whole model symbolically traced.

    Returns:
        GraphModule: Modified GraphModule that has stochastic residual connections.
    """
    if not isinstance(gm, GraphModule):
        raise ValueError(
            f'Input to apply_stochastic_residual should be an instance of GraphModule. Received {type(gm)}',
        )
    all_tags, count = _tag_residual_nodes(gm)
    split_gm = split_by_tags(gm, all_tags)
    assert isinstance(split_gm, GraphModule)
    for node in split_gm.graph.nodes:
        if node.op != 'call_module':
            continue

        main_submod, residual_submod, idx = _get_residual_modules(split_gm, node)
        if main_submod:
            residual_st_instance = BlockStochasticModule(main_submod, residual_submod, drop_rate)
            split_gm.add_submodule(f'resi_st_{idx}', residual_st_instance)  # type: ignore
            _replace_residual_pattern(split_gm, node, f'resi_st_{idx}', residual_submod is not None)
    split_gm.recompile()
    return split_gm, count


def _can_linears_be_fused(linear_nodes: list[Node], all_modules: Mapping[str, nn.Module]) -> bool:
    """Check if all the linears have bias."""
    # Forcing node.target to str is fine here as we are dealing with nn.Modules
    # and their target is a str.
    bias = all_modules[str(linear_nodes[0].target)].bias is None

    return all(bias == (all_modules[str(node.target)].bias is None) for node in linear_nodes)


def _create_fused_linear(
    linear_nodes: list[Node],
    all_modules: Mapping[str, nn.Module],
    keep_weights: bool = False,
) -> tuple[nn.Module, list[int]]:
    """Check if the linears can be fused.

    If the linears can be fused, create a fused nn.Linear instance and return it.
    """
    if keep_weights:
        raise NotImplementedError('This feature is currently not implemented.')

    assert len(linear_nodes) > 1, 'There should be at least 2 linears for fusion'
    out_features = []
    in_features = all_modules[str(linear_nodes[0].target)].in_features
    bias = all_modules[str(linear_nodes[0].target)].bias is not None

    for node in linear_nodes:
        out_features.append(all_modules[str(node.target)].out_features)
        assert in_features == all_modules[str(node.target)].in_features, 'mismatch in number of input features'
        assert bias == (all_modules[str(node.target)].bias is not None), 'mismatch in bias'

    return nn.Linear(in_features, sum(out_features), bias=bias), out_features  # type: ignore


def fuse_parallel_linears(gm: GraphModule, keep_weights: bool = False) -> GraphModule:
    """If there are parallel linears in the model, fuse them together.

    .. rubric:: Example

    .. testsetup::

        import torch
        import torch.nn as nn
        from torch.fx import symbolic_trace
        from composer.utils.fx_utils import count_op_instances, fuse_parallel_linears

    .. doctest::

        >>> class M(nn.Module):
        ...   def __init__(self):
        ...     super().__init__()
        ...     self.fc1 = nn.Linear(64, 64)
        ...     self.fc2 = nn.Linear(64, 64)
        ...   def forward(self, x):
        ...     y = self.fc1(x)
        ...     z = self.fc2(x)
        ...     return y + z
        >>> module = M()
        >>> traced = symbolic_trace(module)
        >>> count_op_instances(traced, nn.Linear)
        2
        >>> gm = fuse_parallel_linears(traced)
        >>> count_op_instances(traced, nn.Linear)
        1

    Arguments:
        gm (GraphModule): The source FX-traced graph.

    Returns:
        GraphModule: Modified GraphModule with parallel linears fused.
    """
    all_modules: dict[str, nn.Module] = dict(gm.named_modules())
    fused_count = 0
    for node in gm.graph.nodes:
        # There could be more than two parallel linears
        linears_to_fuse = []

        # Check all the users of current node and collect all linear layers
        for user in list(node.users):
            if user.op == 'call_module' and isinstance(user.target,
                                                       str) and isinstance(all_modules[user.target], nn.Linear):
                linears_to_fuse.append(user)

        # Fuse if there are more than 1 parallel linear layers
        if len(linears_to_fuse) > 1 and _can_linears_be_fused(linears_to_fuse, all_modules):
            lin, out_features = _create_fused_linear(linears_to_fuse, all_modules, keep_weights)
            gm.add_submodule(f'fused_linear_{fused_count}', lin)  # type: ignore
            with gm.graph.inserting_after(node):
                fused_node = gm.graph.call_module(f'fused_linear_{fused_count}', args=(node,))
            # insert the split node
            with gm.graph.inserting_after(fused_node):
                kwargs = {'split_size_or_sections': out_features, 'dim': -1}
                split_node = gm.graph.call_function(torch.split, args=(fused_node,), kwargs=kwargs)

            insert_point = split_node
            for idx, lin_node in enumerate(linears_to_fuse):
                with gm.graph.inserting_after(insert_point):
                    split_item = gm.graph.call_function(operator.getitem, (split_node, idx), {})
                lin_node.replace_all_uses_with(split_item)
                insert_point = split_item
                gm.graph.erase_node(lin_node)
            fused_count += 1
            gm.graph.lint()

    if fused_count > 0:
        gm.recompile()
    return gm



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Profiler Schedules."""

from typing import Callable

from composer.core.state import State
from composer.profiler.profiler_action import ProfilerAction

__all__ = ['cyclic_schedule']


def cyclic_schedule(
    skip_first: int = 0,
    wait: int = 0,
    warmup: int = 1,
    active: int = 4,
    repeat: int = 1,
) -> Callable[[State], ProfilerAction]:
    """Profiler schedule function for a cyclic profiling window.

    This function returns a schedule function that uses a cyclic profiling window. The resulting function can be
    passed as the ``prof_schedule`` argument to the :class:`.Trainer`.

    The cyclic window skips the first ``skip_first`` + ``resumption_batch_idx`` batches in every epoch.
    ``resumption_batch_idx`` is accessed from state.profiler. It is the ``state.timestamp.batch_in_epoch``
    when resuming training.  Then, it performs a cycle of skipping ``wait`` batches, warming up for ``warmup``
    batches, and recording ``active`` batches. It repeats this cycle up to ``repeat`` times per epoch (or
    for the entire epoch, if ``repeat`` is 0). This logic repeats every epoch.

    Args:
        skip_first (int, optional): Number of batches to skip profiling at epoch start.  Defaults to ``0``.
        wait (int, optional): For each profiling cycle, number of batches to skip at the beginning of the cycle.
            Defaults to ``0``.
        warmup (int, optional): For each profiling cycle, number of batches to be in the warmup state after skipping
            ``wait`` batches. Defaults to ``1``.
        active (int, optional): For each profiling cycle, number of batches to record after warming up.  Defaults to ``4``.
        repeat (int, optional): Number of profiling cycles to perform per epoch. Set to ``0`` to record the entire epoch.
            Defaults to ``1``.

    Returns:
        (State -> ProfilerAction): A ``prof_schedule`` for the :class:`.Trainer`.
    """

    def schedule(state: State):
        # do wait, then warump, then active, up to repeat times per cycle
        cycle_len = wait + warmup + active
        batch_idx = int(state.timestamp.batch_in_epoch)
        if state.profiler is not None:
            skip_first_after_resumption = skip_first + state.profiler.resumption_batch_idx
        else:
            skip_first_after_resumption = skip_first
        if batch_idx < skip_first_after_resumption:
            return ProfilerAction.SKIP
        if repeat != 0 and batch_idx >= cycle_len * repeat + skip_first_after_resumption:
            # exhausted the repeat
            return ProfilerAction.SKIP
        position_in_cycle = (batch_idx - skip_first_after_resumption) % cycle_len
        if position_in_cycle < wait:
            return ProfilerAction.SKIP
        if position_in_cycle < wait + warmup:
            return ProfilerAction.WARMUP
        is_last_batch_in_epoch = state.dataloader_len is not None and state.timestamp.batch_in_epoch == state.dataloader_len - 1
        if position_in_cycle == cycle_len - 1 or is_last_batch_in_epoch:
            return ProfilerAction.ACTIVE_AND_SAVE
        return ProfilerAction.ACTIVE

    return schedule



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Callback to save checkpoints during training."""

from __future__ import annotations

import logging
import os
import pathlib
import shutil
import tempfile
import textwrap
from pathlib import Path
from typing import Any, Callable, Optional, Union

from composer.core import Callback, Event, State, Time, Timestamp
from composer.loggers import Logger, MLFlowLogger
from composer.utils import (
    FORMAT_NAME_WITH_DIST_AND_TIME_TABLE,
    FORMAT_NAME_WITH_DIST_TABLE,
    PartialFilePath,
    checkpoint,
    create_interval_scheduler,
    create_symlink_file,
    dist,
    ensure_folder_has_no_conflicting_files,
    format_name_with_dist,
    format_name_with_dist_and_time,
    is_model_deepspeed,
    partial_format,
)
from composer.utils.compression import get_compressor, is_compressed_pt
from composer.utils.object_store.mlflow_object_store import MLFLOW_EXPERIMENT_ID_FORMAT_KEY, MLFLOW_RUN_ID_FORMAT_KEY

log = logging.getLogger(__name__)

__all__ = ['CheckpointSaver']

_TORCH_DISTRIBUTED_CHECKPOINTS_METADATA_FILENAME = '.metadata'


class CheckpointSaver(Callback):  # noqa: D101
    __doc__ = f"""Callback to save checkpoints.

    .. note::

        If the ``folder`` argument is specified when constructing the :class:`.Trainer`, then the :class:`.CheckpointSaver`
        callback need not be constructed manually. However, for advanced checkpointing use cases
        (such as saving a weights-only checkpoint at one interval and the full training state
        at another interval), instance(s) of this :class:`.CheckpointSaver` callback can be specified in the
        ``callbacks`` argument of the :class:`.Trainer`, as shown in the example below.

    Example

    .. testsetup::

        from composer.callbacks.checkpoint_saver import CheckpointSaver

    .. doctest::

        >>> trainer = Trainer(..., callbacks=[
        ...     CheckpointSaver(
        ...         folder='{{run_name}}/checkpoints',
        ...         filename="ep{{epoch}}-ba{{batch}}-rank{{rank}}",
        ...         latest_filename="latest-rank{{rank}}",
        ...         save_interval="1ep",
        ...         weights_only=False,
        ...     )
        ... ])

    Args:
        folder (str, optional): Format string for the save_folder where checkpoints will be saved.
            Default: ``'{{run_name}}/checkpoints'``.

            The following format variables are available:

            {textwrap.indent(FORMAT_NAME_WITH_DIST_TABLE, prefix='            ')}

            .. note::

                When training with multiple devices (i.e. GPUs), ensure that ``'{{rank}}'`` appears in the format.
                Otherwise, multiple processes may attempt to write to the same file.

        filename (str, optional): A format string describing how to name checkpoints.
            Default: ``'ep{{epoch}}-ba{{batch}}-rank{{rank}}.pt'``.

            Checkpoints will be saved approximately to ``{{folder}}/{{filename.format(...)}}``.

            The following format variables are available:

            {textwrap.indent(FORMAT_NAME_WITH_DIST_AND_TIME_TABLE, prefix='            ')}


            .. note::

                *   By default, only the rank zero process will save a checkpoint file.

                *   When using DeepSpeed, each rank will save a checkpoint file in tarball format. DeepSpeed
                    requires tarball format, as it saves model and optimizer states in separate files.
                    Ensure that ``'{{rank}}'`` appears within the ``filename``. Otherwise, multiple ranks
                    may attempt to write to the same file(s), leading to corrupted checkpoints. If no tarball file
                    extension is specified, ``'.tar'`` will be used.

                *   To write to compressed tar files (regardless of whether DeepSpeed is enabled), set the file
                    extension to ``'.tar.gz'``, ``'.tgz'``, ``'.tar.bz2'``, or ``'.tar.lzma'`` (depending on the
                    desired compression algorithm).

                *   To write to compressed pt files (when DeepSpeed is disabled), set the file extension to
                    ``'.pt.bz2'``, ``'.pt.gz'``, ``'.pt.lz4'``, ``'.pt.lzma'``, ``'.pt.lzo'``, ``'.pt.xz'``,
                    ``'.pt.zst'``
                    (depending on the desired algorithm). You must have the corresponding CLI tool installed.
                    ``lz4`` is a good choice for a modest space saving while being very fast to compress.

            .. warning::

                Using compression will block the training loop while checkpoints are being compressed and the
                compressibility of checkpoints can vary significantly depending on your setup. As such, we
                recommend saving checkpoints without compression by default.

                If you have the ``lz4`` command available on your system, you may want to try saving as ``.pt.lz4``
                as the overhead is minimal (usually less than a second) and the saved space can sometimes
                be significant (1% - 40%).

            Consider the following scenario where:

            *   The :attr:`~.State.run_name` is ``'awesome-training-run'``
            *   The default ``folder='{{run_name}}/checkpoints'`` is used.
            *   The default ``name='ep{{epoch}}-ba{{batch}}-rank{{rank}}'`` is used.
            *   The current epoch count is ``1``.
            *   The current batch count is ``42``.

            When DeepSpeed is not being used, the rank zero process will save the checkpoint to
            ``"awesome-training-run/checkpoints/ep1-ba42-rank0"``.

            When DeepSpeed is being used, each rank (process) will save checkpoints to::

                awesome-training-run/checkpoints/ep1-ba42-rank0.tar
                awesome-training-run/checkpoints/ep1-ba42-rank1.tar
                awesome-training-run/checkpoints/ep1-ba42-rank2.tar
                ...

        remote_file_name (str, optional): Format string for the checkpoint's remote file name.
            Default: ``"{{run_name}}/checkpoints/ep{{epoch}}-ba{{batch}}-rank{{rank}}"``.

            After the checkpoint is saved, it will be periodically uploaded.
            The remote file name will be determined by this format string.

            .. seealso:: :doc:`Uploading Files</trainer/file_uploading>` for notes for file uploading.

            The same format variables for ``filename`` are available.

            Leading slashes (``'/'``) will be stripped.

            To disable uploading checkpoints, set this parameter to ``None``.
        latest_filename (str, optional): A format string for a symlink which points to the last saved checkpoint.
            Default: ``'latest-rank{{rank}}.pt'``.

            Symlinks will be created approximately at ``{{folder}}/{{latest_filename.format(...)}}``.

            The same format variables as for ``name`` are available.

            To disable symlinks, set this parameter to ``None``.

            Consider the following scenario, where:

            *   The :attr:`~.State.run_name` is 'awesome-training-run'
            *   The default ``folder='{{run_name}}/checkpoints'`` is used.
            *   The default ``name='ep{{epoch}}-ba{{batch}}-rank{{rank}}'`` is used.
            *   The default ``latest_filename='latest-rank{{rank}}'`` is used.
            *   The current epoch count is ``1``.
            *   The current batch count is ``42``.

            When DeepSpeed is not being used, the rank zero process will save the checkpoint to
            ``'awesome-training-run/checkpoints/ep1-ba42-rank0'``,
            and a symlink will be created at
            ``'awesome-training-run/checkpoints/latest-rank0' -> 'awesome-training-run/checkpoints/ep1-ba42-rank0'``

            When DeepSpeed is being used, each rank (process) will save checkpoints to::

                awesome-training-run/checkpoints/ep1-ba42-rank0.tar
                awesome-training-run/checkpoints/ep1-ba42-rank1.tar
                awesome-training-run/checkpoints/ep1-ba42-rank2.tar
                ...

            Corresponding symlinks will be created at::

                awesome-training-run/checkpoints/latest-rank0.tar -> awesome-training-run/checkpoints/ep1-ba42-rank0.tar
                awesome-training-run/checkpoints/latest-rank1.tar -> awesome-training-run/checkpoints/ep1-ba42-rank1.tar
                awesome-training-run/checkpoints/latest-rank2.tar -> awesome-training-run/checkpoints/ep1-ba42-rank2.tar
                ...
        latest_remote_file_name (str, optional): Format string for the checkpoint's latest symlink remote file name.
            Default: ``'{{run_name}}/checkpoints/latest-rank{{rank}}"``.

            Whenever a new checkpoint is saved, a symlink is created or updated to point to the latest checkpoint's ``remote_file_name``.
            The remote file name will be determined by this format string. This parameter has no effect if ``latest_filename`` or ``remote_file_name`` is ``None``.

            .. seealso:: :doc:`Uploading Files</trainer/file_uploading>` for notes for file uploading.

            The same format variables for ``filename`` are available.

            Leading slashes (``'/'``) will be stripped.

            To disable symlinks in logger, set this parameter to ``None``.

        overwrite (bool, optional): Whether existing checkpoints should be overridden.
            If ``False`` (the default), then the ``folder`` must not exist or must not contain checkpoints which may conflict
            with the current run. Default: ``False``.

        save_interval (Time | str | int | (State, Event) -> bool): A :class:`.Time`, time-string, integer (in epochs),
            or a function that takes (state, event) and returns a boolean whether a checkpoint should be saved.

            If an integer, checkpoints will be saved every n epochs.
            If :class:`.Time` or a time-string, checkpoints will be saved according to this interval.

            .. seealso:: :func:`.checkpoint_periodically`

            If a function, then this function should take two arguments (:class:`.State`, :class:`.Event`).
            The first argument will be the current state of the trainer, and the second argument will be
            be :attr:`.Event.BATCH_CHECKPOINT` or :attr:`.Event.EPOCH_CHECKPOINT` (depending on the current training
            progress). It should return ``True`` if a checkpoint should be saved given the current state and
            event.

        num_checkpoints_to_keep (int, optional): The number of checkpoints to keep locally. The oldest checkpoints
            are removed first. Set to ``-1`` to keep all checkpoints locally. Default: ``-1``.

            Checkpoints will be removed after they have been uploaded. For example, when this callback
            is used in conjunction with the :class:`.RemoteUploaderDownloader`, set this
            parameter to ``0`` to immediately delete checkpoints from the local disk after they have been uploaded to
            the object store.

            This parameter only controls how many checkpoints are kept locally; checkpoints are not deleted from
            remote file systems.

        weights_only (bool): If ``True``, save only the model weights instead of the entire training state.
            This parameter must be ``False`` when using DeepSpeed. Default: ``False``.

        ignore_keys (list[str] | (dict) -> None, optional): A list of paths for the ``state_dict`` of the checkpoint,
            which, when provided, will be ignored from the state_dict before a checkpoint is saved. Each path is a list
            of strings specifying the keys to index into ``state_dict`` joined together with `/` as a separator (as PyTorch
            uses `.` in parameter names). If a prefix is provided, all children are also ignored (see Example 2).
            See :mod:`composer.core.state` for the structure of state_dict.

            Example 1: ``save_ignore_keys = ["state/model/layer1.weights", "state/model/layer1.bias"]`` would ignore
            layer 1 weights and bias.

            Example 2: ``save_ignore_keys = ["state/model/*"]`` would ignore the entire model, which would have the same
            effect as the previous example if there was only 1 layer.

            Example 3: ``save_ignore_keys = ["state/model/layer*.weights"]`` would ignore all weights in the model.

            Example 4: ``save_ignore_keys = ["state/rank_zero_seed", "rng"]`` would reset all randomness when
            saving the checkpoint.

            If a callable, it should take one argument which is the state_dict. The callable is free to arbitrarily modify
            the state_dict before it is loaded.

            (default: ``None``)

    Attributes:
        saved_checkpoints (list[tuple[Timestamp, list[pathlib.Path]]]): The checkpoint timestamps and filepaths.

            This list contains tuples of the save timestamp and the checkpoint filepaths.
            This list will have at most ``num_checkpoints_to_keep`` entries. The latest checkpoint
            will be at the end.

            .. note::

                When using DeepSpeed, the index of a filepath in each list corresponds to the global rank of
                the process that wrote that file. Each filepath is valid only on the process's (rank's) node.

                Otherwise, when not using DeepSpeed, each sub-list will contain only one filepath since only rank zero
                saves checkpoints.
    """

    def __init__(
        self,
        folder: Union[str, pathlib.Path] = '{run_name}/checkpoints',
        filename: Union[str, pathlib.Path] = 'ep{epoch}-ba{batch}-rank{rank}.pt',
        remote_file_name: Optional[Union[str, pathlib.Path]
                                  ] = ('{run_name}/checkpoints/'
                                       'ep{epoch}-ba{batch}-rank{rank}.pt'),
        latest_filename: Optional[Union[str, pathlib.Path]] = 'latest-rank{rank}.pt',
        latest_remote_file_name: Optional[Union[str, pathlib.Path]] = '{run_name}/checkpoints/latest-rank{rank}.pt',
        save_interval: Union[Time, str, int, Callable[[State, Event], bool]] = '1ep',
        *,
        overwrite: bool = False,
        num_checkpoints_to_keep: int = -1,
        weights_only: bool = False,
        ignore_keys: Optional[Union[list[str], Callable[[dict], None]]] = None,
    ):
        folder = str(folder)
        filename = str(filename)
        remote_file_name = str(remote_file_name) if remote_file_name is not None else None
        latest_filename = str(latest_filename) if latest_filename is not None else None
        latest_remote_file_name = str(latest_remote_file_name) if latest_remote_file_name is not None else None

        # want to fail early if a required CLI tool is missing to ensure no training time is wasted
        for name in [filename, remote_file_name, latest_filename, latest_remote_file_name]:
            if name is not None and is_compressed_pt(name):
                get_compressor(name).check_exists()

        if not callable(save_interval):
            save_interval = create_interval_scheduler(save_interval)
        self.save_interval = save_interval
        self.last_checkpoint_batch: Optional[Time] = None

        self.folder = folder

        self.filename = PartialFilePath(filename.lstrip('/'), folder)
        self.latest_filename = PartialFilePath(latest_filename.lstrip('/'), folder) if latest_filename else None
        self.remote_file_name = PartialFilePath(remote_file_name) if remote_file_name else None
        self.latest_remote_file_name = PartialFilePath(latest_remote_file_name) if latest_remote_file_name else None

        self.overwrite = overwrite
        self.saved_checkpoints: list[str] = []
        self.all_saved_checkpoints_to_timestamp: dict[str, Timestamp] = {}
        self.num_checkpoints_to_keep = num_checkpoints_to_keep
        self.weights_only = weights_only
        self.ignore_keys = ignore_keys

        self.start_batch = None

    def init(self, state: State, logger: Logger) -> None:
        # If MLFlowLogger is being used, format MLFlow-specific placeholders in the save folder and paths.
        # Assumes that MLFlowLogger comes before CheckpointSaver in the list of loggers.
        for destination in logger.destinations:
            if isinstance(destination, MLFlowLogger):
                mlflow_format_kwargs = {
                    MLFLOW_EXPERIMENT_ID_FORMAT_KEY: destination._experiment_id,
                    MLFLOW_RUN_ID_FORMAT_KEY: destination._run_id,
                }
                self.folder = partial_format(self.folder, **mlflow_format_kwargs)

                self.filename.folder = self.folder
                if self.latest_filename is not None:
                    self.latest_filename.folder = self.folder

                # The remote paths have the placeholders in their filename rather than folder
                if self.remote_file_name is not None:
                    self.remote_file_name.filename = partial_format(
                        self.remote_file_name.filename,
                        **mlflow_format_kwargs,
                    )
                if self.latest_remote_file_name is not None:
                    self.latest_remote_file_name.filename = partial_format(
                        self.latest_remote_file_name.filename,
                        **mlflow_format_kwargs,
                    )

                break

        folder = format_name_with_dist(self.folder, state.run_name)
        os.makedirs(folder, exist_ok=True)

    def fit_start(self, state: State, logger: Logger) -> None:
        if not self.overwrite:
            # checks that save_folder contains no files with a timestamp after the current timestamp,
            # which has potential for future conflicts.
            folder = format_name_with_dist(self.folder, state.run_name)
            ensure_folder_has_no_conflicting_files(folder, self.filename.filename, state.timestamp)

        dist.barrier()  # holds all ranks until folder check is done

        if is_model_deepspeed(state.model) and self.weights_only:
            raise NotImplementedError('weights_only=True is not supported when using DeepSpeed.')

        self.start_batch = state.timestamp.batch

    def batch_checkpoint(self, state: State, logger: Logger):
        assert callable(self.save_interval)
        if self.save_interval(state, Event.BATCH_CHECKPOINT) and self.last_checkpoint_batch != state.timestamp.batch:
            self._save_checkpoint(
                state,
                logger,
            )

    def epoch_checkpoint(self, state: State, logger: Logger):
        assert callable(self.save_interval)
        if self.save_interval(state, Event.EPOCH_CHECKPOINT) and self.last_checkpoint_batch != state.timestamp.batch:
            self._save_checkpoint(
                state,
                logger,
            )

    def iteration_checkpoint(self, state: State, logger: Logger):
        assert callable(self.save_interval)
        if (
            self.save_interval(state, Event.ITERATION_CHECKPOINT) and
            self.last_checkpoint_batch != state.timestamp.batch
        ):
            self._save_checkpoint(
                state,
                logger,
            )

    def state_dict(self) -> dict[str, Any]:
        state_dict = {}

        all_checkpoints = []
        for save_filename, timestamp in self.all_saved_checkpoints_to_timestamp.items():
            all_checkpoints.append((save_filename, timestamp.state_dict()))

        state_dict['all_saved_checkpoints_to_timestamp'] = all_checkpoints
        return state_dict

    def load_state_dict(self, state: dict[str, Any]):
        if 'all_saved_checkpoints_to_timestamp' in state:
            for (save_filename, timestamp_state) in state['all_saved_checkpoints_to_timestamp']:
                load_timestamp = Timestamp()
                load_timestamp.load_state_dict(timestamp_state)
                self.all_saved_checkpoints_to_timestamp[save_filename] = load_timestamp

    def _save_checkpoint(self, state: State, logger: Logger):
        self.last_checkpoint_batch = state.timestamp.batch

        is_deepspeed = is_model_deepspeed(state.model)

        if is_deepspeed and '{rank}' not in self.filename.filename:
            raise ValueError(f'Save filename {self.filename.filename} must have {{rank}} for deepspeed.')

        # save the checkpoint to the filename
        filename_with_placeholders = self.filename.format(state, is_deepspeed, keep_placeholders=True)
        save_filename = checkpoint.get_save_filename(state, filename_with_placeholders)
        # Store before saving so state_dict in checkpoint has reference to latest checkpoint (itself)
        self.all_saved_checkpoints_to_timestamp[save_filename] = state.timestamp

        saved_path = checkpoint.save_checkpoint(
            state=state,
            filename=filename_with_placeholders,
            weights_only=self.weights_only,
            ignore_keys=self.ignore_keys,
        )
        log.debug(f'Checkpoint locally saved to {saved_path}')

        if not saved_path:  # not all ranks save
            return

        metadata_local_file_path = None
        if dist.get_global_rank() == 0 and state.fsdp_sharded_state_dict_enabled:
            metadata_local_file_path = format_name_with_dist_and_time(
                os.path.join(Path(saved_path).parent, _TORCH_DISTRIBUTED_CHECKPOINTS_METADATA_FILENAME),
                state.run_name,
                state.timestamp,
            )

        if self.latest_filename is not None and self.num_checkpoints_to_keep != 0:
            symlink = self.latest_filename.format(state, is_deepspeed)
            os.makedirs(os.path.dirname(symlink), exist_ok=True)
            try:
                os.remove(symlink)
            except FileNotFoundError:
                pass
            # Sharded checkpoints for torch >2.0 use directories not files for load_paths
            if state.fsdp_sharded_state_dict_enabled:
                src_path = str(pathlib.Path(saved_path).parent)
            else:
                src_path = saved_path
            this_rank_saves_symlinks = dist.get_global_rank() == 0 or not state.fsdp_sharded_state_dict_enabled
            if this_rank_saves_symlinks:
                os.symlink(os.path.relpath(src_path, os.path.dirname(symlink)), symlink)

        # if remote file name provided, upload the checkpoint
        if self.remote_file_name is not None:
            if state.fsdp_sharded_state_dict_enabled:
                remote_file_name = self.remote_file_name.format(
                    state,
                    is_deepspeed,
                    keep_placeholders=True,
                ).lstrip('/')
                assert state.fsdp_config is not None
                remote_prefix = state.fsdp_config.sharded_ckpt_prefix_dir
                assert remote_prefix is not None
                ckpt_filename = checkpoint._TORCH_DISTRIBUTED_CHECKPOINTS_FILENAME
                remote_file_name = os.path.join(pathlib.Path(remote_file_name).parent, remote_prefix, ckpt_filename)
                remote_file_name = format_name_with_dist_and_time(remote_file_name, state.run_name, state.timestamp)
                # Upload metadata file.
                # The metadata file contains info related to which shards are saved where.
                if dist.get_global_rank() == 0 and state.fsdp_sharded_state_dict_enabled:
                    metadata_remote_file_name = format_name_with_dist_and_time(
                        os.path.join(Path(remote_file_name).parent, _TORCH_DISTRIBUTED_CHECKPOINTS_METADATA_FILENAME),
                        state.run_name,
                        state.timestamp,
                    )
                    assert metadata_local_file_path is not None
                    logger.upload_file(
                        remote_file_name=metadata_remote_file_name,
                        file_path=metadata_local_file_path,
                        overwrite=self.overwrite,
                    )
            else:
                remote_file_name = self.remote_file_name.format(
                    state,
                    is_deepspeed,
                ).lstrip('/')

            log.debug(f'Uploading checkpoint to {remote_file_name}')
            try:
                logger.upload_file(remote_file_name=remote_file_name, file_path=saved_path, overwrite=self.overwrite)
            except FileExistsError as e:
                raise FileExistsError(
                    f'Uploading checkpoint failed with error: {e}. overwrite was set to {self.overwrite}. To overwrite checkpoints with Trainer, set save_overwrite to True.',
                ) from e

            # symlinks stay the same with sharded checkpointing
            if self.latest_remote_file_name is not None:
                symlink_name = self.latest_remote_file_name.format(
                    state,
                    is_deepspeed,
                ).lstrip('/') + '.symlink'

                # create and upload a symlink file
                with tempfile.TemporaryDirectory() as tmpdir:
                    symlink_filename = os.path.join(tmpdir, 'latest.symlink')
                    # Sharded checkpoints for torch >2.0 use directories not files for load_paths
                    if state.fsdp_sharded_state_dict_enabled:
                        src_path = str(pathlib.Path(remote_file_name).parent)
                    else:
                        src_path = remote_file_name
                    log.debug(f'Creating symlink file {symlink_filename} -> {src_path}')
                    this_rank_saves_symlinks = dist.get_global_rank() == 0 or not state.fsdp_sharded_state_dict_enabled
                    if this_rank_saves_symlinks:
                        create_symlink_file(src_path, symlink_filename)
                        logger.upload_file(
                            remote_file_name=symlink_name,
                            file_path=symlink_filename,
                            overwrite=True,
                        )

        self.saved_checkpoints.append(saved_path)

        if self.num_checkpoints_to_keep >= 0:
            self._rotate_checkpoints(sharding_enabled=state.fsdp_sharded_state_dict_enabled)

    def _rotate_checkpoints(self, sharding_enabled: bool = False):

        while len(self.saved_checkpoints) > self.num_checkpoints_to_keep:
            prefix_dir = None
            checkpoint_to_delete = self.saved_checkpoints.pop(0)
            prefix_dir = str(Path(checkpoint_to_delete).parent)
            if not sharding_enabled:
                os.remove(checkpoint_to_delete)
            else:
                if dist.get_global_rank() == 0:
                    shutil.rmtree(prefix_dir)



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Profiler to collect :mod:`torch` performance metrics during training."""

from __future__ import annotations

import json
import logging
import os
import textwrap
from typing import TYPE_CHECKING, Optional, OrderedDict

import torch.cuda
import torch.profiler
from packaging import version
from torch.profiler.profiler import ProfilerAction as TorchProfilerAction

from composer.core.callback import Callback
from composer.loggers import Logger
from composer.profiler.profiler_action import ProfilerAction
from composer.utils import (
    FORMAT_NAME_WITH_DIST_AND_TIME_TABLE,
    FORMAT_NAME_WITH_DIST_TABLE,
    dist,
    ensure_folder_is_empty,
    format_name_with_dist,
    format_name_with_dist_and_time,
)

if TYPE_CHECKING:
    from composer.core import State

__all__ = ['TorchProfiler']

log = logging.getLogger(__name__)


class TorchProfiler(Callback):  # noqa: D101
    __doc__ = f"""Profile the execution using the :class:`PyTorch Profiler <torch.profiler.profile>`.

    Profiling results are stored in TensorBoard format in the directory specified by ``folder``.

    .. note::

        The Composer :class:`.Trainer` automatically creates an instance of this
        :class:`.TorchProfiler` callback whenever any of the PyTorch Profiler arguments
        (``torch_prof_record_shapes``, ``torch_prof_profile_memory``, ``torch_prof_with_stack``, or
        ``torch_prof_with_flops``) are enabled.

        When using the Composer :class:`.Trainer`, one does not need to directly create an
        instance of this :class:`.TorchProfiler` callback.


    To view profiling results, run::

        pip install tensorboard torch_tb_profiler
        tensorboard --logdir path/to/torch/trace_folder

    .. note::

        See :doc:`profiler` for additional usage details on the :class:`torch.profiler.profile`.

    .. note::

        Enabling shape and stack tracing results in additional overhead.
        When ``record_shapes=True`` is specified, the profiler will temporarily hold references to tensors which
        may prevent certain optimizations that depend on the reference count and can introduce extra tensor copies.

    Args:
        folder (str, optional): Format string for the folder containing the Torch Profiler trace files.
            Defaults to ``'{{run_name}}/torch_traces'``.

            The following format variables are available:

            {textwrap.indent(FORMAT_NAME_WITH_DIST_TABLE, prefix='            ')}

            For example, if the ``run_name`` is ``'awesome_training_run'``, and the default ``folder`` of
            ``'{{run_name}}/torch_traces'`` is used, Torch Profiler traces will be stored in
            ``'awesome_training_run/torch_traces'``.

        filename (str, optional): A format string describing how to name Torch Profiler trace files.
            Defaults to ``'rank{{rank}}.{{batch}}.pt.trace.json'``.

            At the end of each batch where :meth:`~composer.profiler.Profiler.get_action` returns
            :attr:`~composer.profiler._profiler_action.ProfilerAction.ACTIVE_AND_SAVE`, trace files are saved
            approximately to ``{{folder.format(...)}}/{{filename.format(...)}}``.

            The following format variables are available:

            {textwrap.indent(FORMAT_NAME_WITH_DIST_AND_TIME_TABLE, prefix='            ')}

            Consider the following scenario, where:

            *   The :attr:`~.State.run_name` is ``'awesome-training-run'``.
            *   The default ``trace_folder='{{run_name}}/torch_traces'`` is used.
            *   The default ``name='rank{{rank}}.{{batch}}.pt.trace.json'`` is used.
            *   The current epoch count is ``1``.
            *   The current batch count is ``42``.

            Each rank (process) will save traces to::

                awesome-training-run/torch_traces/ep1-ba42-rank0.pt.trace.json
                awesome-training-run/torch_traces/ep1-ba42-rank1.pt.trace.json
                awesome-training-run/torch_traces/ep1-ba42-rank2.pt.trace.json
                ...

        remote_file_name (str, optional): Format string for a Torch Profiler trace file's remote file name.
            Defaults to ``'{{run_name}}/torch_traces/rank{{rank}}.{{batch}}.pt.trace.json'``.

            Whenever a trace file is saved, it is also uploaded as a file according to this format string.
            The same format variables as for ``filename`` are available.

            .. seealso:: :doc:`Uploading Files</trainer/file_uploading>` for notes for file uploading.

            Leading slashes (``'/'``) will be stripped.

            To disable uploading trace files, set this parameter to ``None``.
        memory_filename (str, optional): A format string describing how to name Torch Profiler memory trace files.
            Defaults to None. An example memory_filename is ``'rank{{rank}}.{{batch}}.pt.trace.memory.html'``.

            At the end of each batch where :meth:`~composer.profiler.Profiler.get_action` returns
            :attr:`~composer.profiler._profiler_action.ProfilerAction.ACTIVE_AND_SAVE`, trace files are saved
            approximately to ``{{folder.format(...)}}/{{memory_filename.format(...)}}``.

            The following format variables are available:

            {textwrap.indent(FORMAT_NAME_WITH_DIST_AND_TIME_TABLE, prefix='            ')}

            Consider the following scenario, where:

            *   The :attr:`~.State.run_name` is ``'awesome-training-run'``.
            *   The default ``trace_folder='{{run_name}}/torch_traces'`` is used.
            *   The default ``name='rank{{rank}}.{{batch}}.pt.trace.memory.html'`` is used.
            *   The current epoch count is ``1``.
            *   The current batch count is ``42``.

            Each rank (process) will save traces to::

                awesome-training-run/torch_traces/ep1-ba42-rank0.pt.trace.memory.html
                awesome-training-run/torch_traces/ep1-ba42-rank1.pt.trace.memory.html
                awesome-training-run/torch_traces/ep1-ba42-rank2.pt.trace.memory.html
                ...

        memory_remote_file_name (str, optional): Format string for a Torch Profiler memory trace file's remote file name.
            Defaults to ``'{{run_name}}/torch_traces/rank{{rank}}.{{batch}}.pt.trace.memory.json'``.

            Whenever a trace file is saved, it is also uploaded as a file according to this format string.
            The same format variables as for ``filename`` are available.

            .. seealso:: :doc:`Uploading Files</trainer/file_uploading>` for notes for file uploading.

            Leading slashes (``'/'``) will be stripped.

            To disable uploading trace files, set this parameter to ``None``.
        overwrite (bool, optional): Whether to override existing Torch Profiler traces. Defaults to False.

            If False, then the trace folder as determined by ``folder`` must be empty.
        use_gzip (bool, optional): Whether to use gzip for the trace. Defaults to False.
            If True, ``'.gz'`` will be appended ``filename`` and ``remote_file_name``
            (if they do not already end in ``'.gz'``).
        record_shapes (bool, optional): Whether to record tensor shapes. Defaults to False.
        profile_memory (bool, optional): Whether to profile memory. Defaults to True.
        with_stack (bool, optional): Whether to record stack info. Defaults to False.
        with_flops (bool, optional): Whether to estimate flops for operators. Defaults to True.
        num_traces_to_keep (int, optional): The number of trace files to keep locally. Defaults to -1.

            If set to -1, then all traces files are kept locally.

            After a trace has been saved and uploaded, the oldest traces are removed until
            ``num_traces_to_keep`` traces remain. This parameter only controls how many traces are kept locally;
            traces are not deleted from remote file systems.

            It can be useful to set this parameter to ``0`` when using a remote file uploader such as the
            :class:`.RemoteUploaderDownloader`. This combination will minimize local
            disk usage by deleting trace files immediately after they have been uploaded to the object store.

    Attributes:
        saved_traces (list[tuple[Timestamp, list[pathlib.Path]]]): The trace timestamps and filepaths.

            This list contains tuples of the save timestamp and the trace filepaths.
            This list will have at most ``num_traces_to_keep`` entries. The latest trace
            will be at the end.

            The index of a filepath in each list corresponds to the global rank of the process that wrote that file.
            Each filepath is valid only on the process's (rank's) node.
    """

    def __init__(
        self,
        folder: str = '{run_name}/torch_traces',
        filename: str = 'rank{rank}.{batch}.pt.trace.json',
        remote_file_name: Optional[str] = '{run_name}/torch_traces/rank{rank}.{batch}.pt.trace.json',
        memory_filename: Optional[str] = None,
        memory_remote_file_name: Optional[str] = (
            '{run_name}/torch_memory_traces/'
            'rank{rank}.{batch}.pt.trace.memory.html'
        ),
        overwrite: bool = False,
        use_gzip: bool = False,
        record_shapes: bool = False,
        profile_memory: bool = True,
        with_stack: bool = False,
        with_flops: bool = True,
        num_traces_to_keep: int = -1,
    ) -> None:
        self.overwrite = overwrite
        self.folder = folder

        if use_gzip:
            if not filename.endswith('.gz'):
                filename += '.gz'
        self.filename = filename

        if use_gzip:
            if remote_file_name is not None and not remote_file_name.endswith('.gz'):
                remote_file_name += '.gz'
        self.remote_file_name = remote_file_name

        if memory_filename is not None:
            assert memory_filename.endswith('.html'), f'memory_filename must end with .html, got {memory_filename}'
        self.memory_filename = memory_filename

        if memory_remote_file_name is not None:
            assert memory_remote_file_name.endswith(
                '.html',
            ), f'memory_remote_file_name must end with .html, got {memory_remote_file_name}'
        self.memory_remote_file_name = memory_remote_file_name

        self.record_shapes = record_shapes
        self.profile_memory = profile_memory
        self.with_stack = with_stack
        self.with_flops = with_flops
        self.num_traces_to_keep = num_traces_to_keep
        self.saved_traces = OrderedDict()
        self.profiler: Optional[torch.profiler.profile] = None

    def init(self, state: State, logger: Logger) -> None:
        if state.profiler is None:
            raise RuntimeError((
                'The Composer Profiler was not enabled, which is required to use the '
                f'{type(self).__name__}. To enable, set the `prof_schedule` argument of the Trainer.'
            ))

        folder_name = format_name_with_dist(self.folder, state.run_name)
        os.makedirs(folder_name, exist_ok=True)
        if not self.overwrite:
            ensure_folder_is_empty(folder_name)

        dist.barrier()

        def scheduler_fn(torch_profiler_step: int) -> TorchProfilerAction:
            del torch_profiler_step  # the torch profiler step is unused. Using the composer timestamp instead.

            assert state.profiler is not None
            composer_profiler_action = state.profiler.schedule(state)
            if composer_profiler_action == ProfilerAction.ACTIVE_AND_SAVE:
                return TorchProfilerAction.RECORD_AND_SAVE
            if composer_profiler_action == ProfilerAction.ACTIVE:
                return TorchProfilerAction.RECORD
            if composer_profiler_action == ProfilerAction.WARMUP:
                return TorchProfilerAction.WARMUP
            assert composer_profiler_action == ProfilerAction.SKIP, f'unexpected action: {composer_profiler_action}'
            return TorchProfilerAction.NONE

        def handler_fn(prof: torch.profiler.profiler.profile):

            assert state.profiler is not None

            timestamp = state.timestamp

            log.info(f'PyTorch Chrome trace profiler enabled: {self.filename if self.filename else False}')
            trace_file_name = os.path.join(
                folder_name,
                format_name_with_dist_and_time(self.filename, run_name=state.run_name, timestamp=timestamp),
            )
            trace_file_dirname = os.path.dirname(trace_file_name)
            if trace_file_dirname:
                os.makedirs(trace_file_dirname, exist_ok=True)
            prof.export_chrome_trace(trace_file_name)
            state.profiler.record_chrome_json_trace_file(trace_file_name)
            if self.remote_file_name is not None:
                trace_remote_file_name = format_name_with_dist_and_time(
                    self.remote_file_name,
                    run_name=state.run_name,
                    timestamp=timestamp,
                )
                trace_remote_file_name = trace_remote_file_name.lstrip('/')
                logger.upload_file(
                    remote_file_name=trace_remote_file_name,
                    file_path=trace_file_name,
                    overwrite=self.overwrite,
                )

            log.info(
                f'PyTorch memory timeline profiler enabled: {self.memory_filename if self.memory_filename else False}',
            )
            if self.memory_filename is not None:
                if version.parse(torch.__version__) > version.parse('2.1.0.dev'):  # type: ignore
                    # memory timeline profiling is only supported in torch v2.1.0-rc1 or higher
                    memory_trace_file_name = os.path.join(
                        folder_name,
                        format_name_with_dist_and_time(
                            self.memory_filename,
                            run_name=state.run_name,
                            timestamp=timestamp,
                        ),
                    )
                    log.debug(f'Saving memory trace to {memory_trace_file_name}')
                    memory_trace_file_dirname = os.path.dirname(memory_trace_file_name)
                    if memory_trace_file_dirname:
                        os.makedirs(memory_trace_file_dirname, exist_ok=True)
                    from composer.profiler.utils import export_memory_timeline_html
                    export_memory_timeline_html(
                        prof,
                        memory_trace_file_name,
                        torch.cuda.current_device(),  # type: ignore
                    )
                    log.debug(f'Uploaded memory trace to {self.memory_remote_file_name}')
                    if self.memory_remote_file_name is not None:
                        memory_trace_remote_file_name = format_name_with_dist_and_time(
                            self.memory_remote_file_name,
                            run_name=state.run_name,
                            timestamp=timestamp,
                        )
                        memory_trace_remote_file_name = memory_trace_remote_file_name.lstrip('/')
                        log.debug(
                            f'Uploading memory trace to {memory_trace_remote_file_name} from {memory_trace_file_name}',
                        )
                        logger.upload_file(
                            remote_file_name=memory_trace_remote_file_name,
                            file_path=memory_trace_file_name,
                            overwrite=self.overwrite,
                        )
                else:
                    log.warning('Memory timeline is supported after PyTorch 2.1.0. Skipping memory trace.')

            if self.num_traces_to_keep >= 0:
                while len(self.saved_traces) > self.num_traces_to_keep:
                    # self.saved_traces is an ordered dict, so the zeroth item will be the oldest checkpoint
                    timestamp, filepaths = next(iter(self.saved_traces.items()))
                    if dist.get_global_rank() < len(filepaths):
                        # Remove this rank's checkpoint
                        os.remove(filepaths[dist.get_global_rank()])
                    del self.saved_traces[timestamp]

        self.profiler = torch.profiler.profile(
            schedule=scheduler_fn,
            on_trace_ready=handler_fn,
            record_shapes=self.record_shapes,
            profile_memory=self.profile_memory,
            with_stack=self.with_stack,
            with_flops=self.with_flops,
        )
        self.profiler.__enter__()

    def batch_end(self, state: State, logger: Logger) -> None:
        del state, logger  # unused
        assert self.profiler is not None
        self.profiler.add_metadata_json('global_rank', json.dumps(dist.get_global_rank()))
        self.profiler.step()

    def batch_start(self, state: State, logger: Logger) -> None:
        del state  # unused
        assert self.profiler is not None
        logger.log_traces({'profiler/state': self.profiler.current_action.name})

    def close(self, state: State, logger: Logger) -> None:
        del state, logger  # unused
        if self.profiler is not None and self.profiler.profiler is not None:
            log.info(self.profiler.key_averages().table(sort_by='cpu_time_total', row_limit=20))
            if self.profile_memory:
                log.info(self.profiler.key_averages().table(sort_by='self_cpu_memory_usage', row_limit=20))
            if torch.profiler.ProfilerActivity.CUDA in self.profiler.activities:
                log.info(self.profiler.key_averages().table(sort_by='cuda_time_total', row_limit=20))
                if self.profile_memory:
                    log.info(self.profiler.key_averages().table(sort_by='self_cuda_memory_usage', row_limit=20))
            self.profiler.__exit__(None, None, None)
            self.profiler = None



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Profiler Trace Handler."""

from __future__ import annotations

import abc
import pathlib
from typing import TYPE_CHECKING, Union

from composer.core.callback import Callback

if TYPE_CHECKING:
    from composer.core import Timestamp

__all__ = ['TraceHandler']


class TraceHandler(Callback, abc.ABC):
    """Base class for Composer Profiler trace handlers.

    Subclasses should implement :meth:`process_duration_event`, :meth:`process_instant_event`,
    :meth:`process_counter_event`, and :meth:`process_chrome_json_trace_file` to record trace events.

    Since :class:`TraceHandler` subclasses :class:`.Callback`, a trace handler can run on any
    :class:`.Event` (such as on :attr:`.Event.INIT` to open files or on :attr:`.Event.BATCH_END` to periodically dump
    data to files) and use :meth:`.Callback.close` to perform any cleanup.
    """

    def process_duration_event(
        self,
        name: str,
        categories: Union[list[str], tuple[str, ...]],
        is_start: bool,
        timestamp: Timestamp,
        wall_clock_time_ns: int,
    ) -> None:
        """Invoked whenever there is a duration event to record.

        This method is called twice for each duration event -- once with ``is_start = True``,
        and then again with ``is_start = False``. Interleaving events are not permitted.
        Specifically, for each event (identified by the ``name``), a call with ``is_start = True`` will be followed
        by a call with ``is_start = False`` before another call with ``is_start = True``.

        Args:
            name (str): The name of the event.
            categories (Union[list[str], tuple[str, ...]]): The categories for the event.
            is_start (bool): Whether the event is a start event or end event.
            timestamp (Timestamp): Snapshot of the training time.
            wall_clock_time_ns (int): The :py:func:`time.time_ns` corresponding to the event.
        """
        del name, categories, is_start, timestamp, wall_clock_time_ns  # unused
        pass

    def process_instant_event(
        self,
        name: str,
        categories: Union[list[str], tuple[str, ...]],
        timestamp: Timestamp,
        wall_clock_time_ns: int,
    ) -> None:
        """Invoked whenever there is an instant event to record.

        Args:
            name (str): The name of the event.
            categories (list[str] | tuple[str, ...]): The categories for the event.
            timestamp (Timestamp): Snapshot of current training time.
            wall_clock_time_ns (int): The :py:func:`time.time_ns` corresponding to the event.
        """
        del name, categories, timestamp, wall_clock_time_ns  # unused
        pass

    def process_counter_event(
        self,
        name: str,
        categories: Union[list[str], tuple[str, ...]],
        timestamp: Timestamp,
        wall_clock_time_ns: int,
        values: dict[str, Union[int, float]],
    ) -> None:
        """Invoked whenever there is an counter event to record.

        Args:
            name (str): The name of the event.
            categories (list[str] | tuple[str, ...]): The categories for the event.
            timestamp (Timestamp): The timestamp.
            wall_clock_time_ns (int): The :py:func:`time.time_ns` corresponding to the event.
            values (dict[str, int | float]): The values corresponding to this counter event.
        """
        del name, categories, timestamp, wall_clock_time_ns, values  # unused
        pass

    def process_chrome_json_trace_file(self, filepath: pathlib.Path) -> None:
        """Invoked when there are events in Chrome JSON format to record.

        See `this document <https://docs.google.com/document/d/1CvAClvFfyA5R-PhYUmn5OOQtYMH4h6I0nSsKchNAySU/preview>`_
        for more information.

        Args:
            filepath (pathlib.Path): The filepath to a Chrome JSON trace file.
        """
        del filepath  # unused
        pass



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Stochastic Weight Averaging (SWA; `Izmailov et al, 2018 <https://arxiv.org/abs/1803.05407>`_) averages model weights
sampled at different times near the end of training.

This leads to better generalization than just using the final trained weights. See the :doc:`Method Card
</method_cards/swa>` for more details.
"""

from composer.algorithms.swa.swa import SWA

__all__ = ['SWA']



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Reference for common types used throughout the composer library.

Attributes:
    Batch (Any): Alias to type Any.
        A batch of data can be represented in several formats, depending on the application.
    JSON (str | float | int | None | list['JSON'] | dict[str, 'JSON']): JSON Data.
    Dataset (torch.utils.data.Dataset[Batch]): Alias for :class:`torch.utils.data.Dataset`.
"""

from __future__ import annotations

from typing import Any, Union

import torch
import torch.utils.data

from composer.utils import StringEnum

__all__ = ['Batch', 'JSON', 'MemoryFormat', 'TrainerMode']

Batch = Any

Dataset = torch.utils.data.Dataset[Batch]

JSON = Union[str, float, int, None, list['JSON'], dict[str, 'JSON']]


class TrainerMode(StringEnum):
    """Enum to represent which mode the Trainer is in.

    Attributes:
        TRAIN: In training mode.
        EVAL: In evaluation mode.
        PREDICT: In predict mode.
    """
    TRAIN = 'train'
    EVAL = 'eval'
    PREDICT = 'predict'


class MemoryFormat(StringEnum):
    """Enum class to represent different memory formats.

    See :class:`torch.torch.memory_format` for more details.

    Attributes:
        CONTIGUOUS_FORMAT: Default PyTorch memory format represnting a tensor allocated with consecutive dimensions
            sequential in allocated memory.
        CHANNELS_LAST: This is also known as NHWC. Typically used for images with 2 spatial dimensions (i.e., Height and
            Width) where channels next to each other in indexing are next to each other in allocated memory. For example, if
            C[0] is at memory location M_0 then C[1] is at memory location M_1, etc.
        CHANNELS_LAST_3D: This can also be referred to as NTHWC. Same as :attr:`CHANNELS_LAST` but for videos with 3
            spatial dimensions (i.e., Time, Height and Width).
        PRESERVE_FORMAT: A way to tell operations to make the output tensor to have the same memory format as the input
            tensor.
    """
    CONTIGUOUS_FORMAT = 'contiguous_format'
    CHANNELS_LAST = 'channels_last'
    CHANNELS_LAST_3D = 'channels_last_3d'
    PRESERVE_FORMAT = 'preserve_format'



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Log memory snapshot during training."""
import logging
import os
import pickle
import warnings
from typing import Optional, Union

import torch.cuda
from packaging import version

from composer import State
from composer.core import Callback, State, Time, TimeUnit
from composer.loggers import Logger
from composer.utils import ensure_folder_is_empty, format_name_with_dist, format_name_with_dist_and_time, parse_uri

log = logging.getLogger(__name__)

__all__ = ['MemorySnapshot']


class MemorySnapshot(Callback):
    """Logs the memory snapshot of the model.

    This callback calls the torch memory snapshot API (see :func:`torch.cuda.memory._snapshot`) to record a model's tensor memory allocation over a user defined interval (only once through time [skip_batches, skip_batches + interval]). This provides a fine-grained GPU memory visualization for debugging GPU OOMs. Captured memory snapshots will show memory events including allocations, frees and OOMs, along with their stack traces over one interval.

    Example:
        .. doctest::

            >>> from composer import Trainer
            >>> from composer.callbacks import MemorySnapshot
            >>> # constructing trainer object with this callback
            >>> trainer = Trainer(
            ...     model=model,
            ...     train_dataloader=train_dataloader,
            ...     eval_dataloader=eval_dataloader,
            ...     optimizers=optimizer,
            ...     max_duration="1ep",
            ...     callbacks=[MemorySnapshot()],
            ... )

    .. note::
        Memory snapshot is only supported for GPU devices.

    Args:
        skip_batches (int, optional): Number of batches to skip before starting recording memory snapshot. Defaults to 1.
        interval (Union[int, str, Time], optional): Time string specifying how long to record the tensor allocation.
            For example, ``interval='3ba'`` means 3 batches are recorded. Default: '3ba'.
        max_entries (int, optional): Maximum number of memory alloc/free events to record. Defaults to 100000.
        folder (str, optional): A format string describing the folder containing the memory snapshot files.
            Defaults to ``'{{run_name}}/torch_traces'``.
        filename (str, optional): A format string describing the prefix used to name the memory snapshot files.
            Defaults to ``'rank{{rank}}.{{batch}}.memory_snapshot'``.
        remote_file_name (str, optional): A format string describing the prefix for the memory snapshot remote file name.
            Defaults to ``'{{run_name}}/torch_traces/rank{{rank}}.{{batch}}.memory_snapshot'``.

            Whenever a trace file is saved, it is also uploaded as a file according to this format string.
            The same format variables as for ``filename`` are available.

            .. seealso:: :doc:`Uploading Files</trainer/file_uploading>` for notes for file uploading.

            Leading slashes (``'/'``) will be stripped.

            To disable uploading trace files, set this parameter to ``None``.
        overwrite (bool, optional): Whether to override existing memory snapshots. Defaults to False.

            If False, then the trace folder as determined by ``folder`` must be empty.
    """

    def __init__(
        self,
        skip_batches: int = 1,
        interval: Union[int, str, Time] = '3ba',
        max_entries: int = 100000,
        folder: str = '{run_name}/torch_traces',
        filename: str = 'rank{rank}.{batch}.memory_snapshot',
        remote_file_name: Optional[str] = '{run_name}/torch_memory_traces',
        overwrite: bool = False,
    ) -> None:
        self.batches_left_to_skip = skip_batches
        # Check that the interval timestring is parsable and convert into time object
        self.interval = Time.from_input(interval, TimeUnit.BATCH)
        self.max_entries = max_entries
        self.folder = folder
        self.folder_name = None
        self.filename = filename
        self.remote_file_name = remote_file_name
        self.overwrite = overwrite
        self._start_time = None
        if remote_file_name:
            self.remote_file_name = remote_file_name
            _, _, self.remote_path_in_bucket = parse_uri(remote_file_name)
        else:
            self.remote_path_in_bucket = None

        if version.parse(torch.__version__.split('.dev')[0]) >= version.parse('2.1.0'):  # type: ignore
            # MemorySnapshot is only supported in torch v2.1.0-rc1 or higher
            self._enabled = True
        else:
            self._enabled = False
            warnings.warn('Memory snapshot is supported after PyTorch 2.1.0. Skipping memory snapshot callback.')

    def init(self, state: State, logger: Logger) -> None:
        if not self._enabled:
            return
        # Not relying on `torch.cuda.is_available()` since the model could be on CPU.
        model_device = next(state.model.parameters()).device

        if model_device.type not in ('cuda', 'meta'):
            warnings.warn(f'The memory snapshot only works on CUDA devices, but the model is on {model_device.type}.')
            self._enabled = False
        else:
            self.folder_name = format_name_with_dist(self.folder, state.run_name)
            os.makedirs(self.folder_name, exist_ok=True)
            if not self.overwrite:
                ensure_folder_is_empty(self.folder_name)

    def batch_start(self, state: State, logger: Logger) -> None:
        if self._enabled and self._start_time is None and self.batches_left_to_skip == 0:
            self.start_record_memory_history()
            self._start_time = state.timestamp.get(self.interval.unit).value

    def batch_end(self, state: State, logger: Logger) -> None:
        if not self._enabled:
            return

        if self.batches_left_to_skip > 0:
            self.batches_left_to_skip -= 1
            return
        assert self._start_time is not None

        if state.timestamp.get(self.interval.unit).value == (self._start_time + self.interval.value):
            self.export_memory_snapshot(state, logger)
            self.stop_record_memory_history()
            self._start_time = None
            self._enabled = False

    def start_record_memory_history(self) -> None:

        log.info('Starting snapshot record_memory_history')
        torch.cuda.memory._record_memory_history(
            'all',  # type: ignore
            max_entries=self.max_entries,
        )

    def stop_record_memory_history(self) -> None:

        log.info('Stopping snapshot record_memory_history')
        torch.cuda.memory._record_memory_history(False)  # type: ignore

    def export_memory_snapshot(self, state: State, logger: Logger) -> None:
        assert self.filename
        assert self.folder_name, 'folder_name must be set in init'
        filename = os.path.join(
            self.folder_name,
            format_name_with_dist_and_time(self.filename, run_name=state.run_name, timestamp=state.timestamp),
        )
        try:
            snapshot_file = filename + '.pickle'
            trace_plot_file = filename + '.html'
            log.info(f'Saving memory snapshot files')

            snapshot = torch.cuda.memory._snapshot()
            # No data was recorded - avoids a `ValueError` in `trace_plot`
            if all(len(t) == 0 for t in snapshot['device_traces']):
                log.info(f'No allocation is recorded in memory snapshot)')
                return

            with open(snapshot_file, 'wb') as fd:
                pickle.dump(snapshot, fd)

            with open(trace_plot_file, 'w+') as fd:
                fd.write(torch.cuda._memory_viz.trace_plot(snapshot))  # type: ignore

            log.info(f'Saved memory snapshot to local files with prefix = {filename}')

            if self.remote_path_in_bucket is not None:
                for f in [snapshot_file, trace_plot_file]:
                    remote_file_name = os.path.join(self.remote_path_in_bucket, os.path.basename(f)).lstrip('/')
                    log.info(f'Uploading memory snapshot to remote: {remote_file_name} from {f}')
                    try:
                        logger.upload_file(remote_file_name=remote_file_name, file_path=f, overwrite=self.overwrite)
                    except FileExistsError as e:
                        raise FileExistsError(
                            f'Uploading memory snapshot failed with error: {e}. overwrite was set to {self.overwrite}. To overwrite memory snapshot with Trainer, set `overwrite` to True.',
                        ) from e
        except Exception as e:
            log.error(f'Failed to capture memory snapshot {e}')



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Loggers to store metrics and checkpoints.

In Composer, algorithms and callbacks can make calls to the :class:`~.logger.Logger`,
which then routes the calls to the appropriate :class:`~.logger_destination.LoggerDestination` instances.
The :class:`~.logger_destination.LoggerDestination` does the actual logging, for example to a file,
or Weights and Biases.

See the :class:`~.logger_destination.LoggerDestination` documentation for an example of how to
define a custom logger and use it when training.
"""

from composer.loggers.cometml_logger import CometMLLogger
from composer.loggers.console_logger import ConsoleLogger
from composer.loggers.file_logger import FileLogger
from composer.loggers.in_memory_logger import InMemoryLogger
from composer.loggers.logger import Logger
from composer.loggers.logger_destination import LoggerDestination
from composer.loggers.mlflow_logger import MLFlowLogger
from composer.loggers.mosaicml_logger import MosaicMLLogger
from composer.loggers.neptune_logger import NeptuneLogger
from composer.loggers.progress_bar_logger import ProgressBarLogger
from composer.loggers.remote_uploader_downloader import RemoteUploaderDownloader
from composer.loggers.slack_logger import SlackLogger
from composer.loggers.tensorboard_logger import TensorboardLogger
from composer.loggers.wandb_logger import WandBLogger

# All needs to be defined properly for sphinx autosummary
__all__ = [
    'Logger',
    'LoggerDestination',
    'FileLogger',
    'InMemoryLogger',
    'NeptuneLogger',
    'ProgressBarLogger',
    'WandBLogger',
    'RemoteUploaderDownloader',
    'TensorboardLogger',
    'CometMLLogger',
    'ConsoleLogger',
    'MLFlowLogger',
    'MosaicMLLogger',
    'SlackLogger',
]



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""The Apple M-series device used for training."""

from __future__ import annotations

from typing import TypeVar

import torch
import torch.cuda.amp
import torch.utils.data
from packaging import version

from composer.devices.device import Device

__all__ = ['DeviceMPS']

T_nnModule = TypeVar('T_nnModule', bound=torch.nn.Module)


class DeviceMPS(Device):
    """Device to support MPS, for training on Apple's M-series chips.

    This class takes no arguments.
    """
    dist_backend = ''
    name = 'mps'

    def __init__(self) -> None:
        if version.parse(torch.__version__) < version.parse('1.12.0'):
            raise RuntimeError('Support for MPS device requires torch >= 1.12.')
        if not torch.backends.mps.is_available():  # type: ignore (version guarded)
            raise RuntimeError('MPS requires MAC OSX >= 12.3')
        if not torch.backends.mps.is_built():  # type: ignore (version guarded)
            raise RuntimeError('torch was not build with MPS support.')

        self._device = torch.device('mps')

    def module_to_device(self, module: T_nnModule) -> T_nnModule:
        return module.to(self._device)

    def tensor_to_device(self, tensor: torch.Tensor) -> torch.Tensor:
        return tensor.to(self._device)



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""MCLI compatible eval client."""
import logging
import os
import time
from http import HTTPStatus

import mcli
import numpy as np

from composer.utils.eval_client.eval_client import EvalClient

__all__ = ['MosaicMLLambdaEvalClient']
log = logging.getLogger(__name__)


class MosaicMLLambdaEvalClient(EvalClient):
    """Utility for creating a client for and invoking an AWS Lambda through MCLI."""

    def __init__(self, backoff: int = 3, num_retries: int = 5) -> None:
        """Checks that the requisite environment variables are in the EvalClient.

        `MOSAICML_ACCESS_TOKEN_ENV_VAR` environment variable must be set to access the platform.
        """
        from composer.loggers.mosaicml_logger import \
            MOSAICML_ACCESS_TOKEN_ENV_VAR  # in-line import to avoid circular import

        if MOSAICML_ACCESS_TOKEN_ENV_VAR not in os.environ:
            raise RuntimeError('Cannot use MosaicML Lambda Client Eval without setting MOSAICML_ACCESS_TOKEN_ENV_VAR.')
        log.debug('Running code eval through MosaicMLLambdaEvalClient.')
        self.backoff = backoff
        self.num_retries = num_retries

    def invoke(self, payload: list[list[list[dict[str, str]]]]) -> list[list[list[bool]]]:
        """Invoke a batch of provided payloads for code evaluations."""
        num_beams = len(payload[0])
        num_tests = [len(generation_payload[0]) for generation_payload in payload]
        cum_tests = (np.cumsum([0] + num_tests[:-1]) * num_beams).tolist()
        test_cases = [
            test_case for generation_payload in payload for beam_payload in generation_payload
            for test_case in beam_payload
        ]
        ret_helper = [False] * len(test_cases)
        for i in range(self.num_retries):
            try:
                ret_helper = mcli.get_code_eval_output(test_cases).data  # pyright: ignore[reportGeneralTypeIssues]
                break
            except mcli.MAPIException as e:
                if e.status >= 500:
                    if i == self.num_retries - 1:
                        log.error(f'Failed to get code eval output after {self.num_retries} retries. Error: {e}')
                    log.warning(f'Failed to get code eval output, retrying in {self.backoff**i} seconds.')
                    time.sleep(self.backoff**i)
                elif e.status == HTTPStatus.UNAUTHORIZED:
                    raise RuntimeError(
                        'Failed to get code eval output due to UNAUTHORIZED error. '
                        'Please ensure you have access to MosaicMLLambdaEvalClient.',
                    ) from e
                else:
                    log.error(f'Failed to get code eval output with unexpected MAPIException. Error: {e}')
                    break
            except TimeoutError as e:
                if i == self.num_retries - 1:
                    log.error(f'Failed to get code eval output after {self.num_retries} retries. Error: {e}')
                log.warning(f'Failed to get code eval output, retrying in {self.backoff**i} seconds.')
                time.sleep(self.backoff**i)
            except Exception as e:
                log.error(f'Failed to get code eval output with unexpected error. Error: {e}')
                break

        ret = []
        for i in range(len(payload)):
            ret_payload = []
            for j in range(num_beams):
                ret_num_beams = []
                for k in range(num_tests[i]):
                    ret_num_beams.append(ret_helper[cum_tests[i] + j * num_tests[i] + k])
                ret_payload.append(ret_num_beams)
            ret.append(ret_payload)
        return ret



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

# Copyright 2023 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Abstract class for utilities that access and run code on serverless eval clients."""

import abc

__all__ = ['EvalClient']


class EvalClient(abc.ABC):
    """Abstract class for implementing eval clients, such as LambdaEvalClient."""

    def invoke(self, payload: list[list[list[dict[str, str]]]]) -> list[list[list[bool]]]:
        """Invoke a provided batch of dictionary payload to the client.

        For code generation, the payload is a list of list of lists of JSONs. The lists are organized in a nested structure, with the outer list
        being grouped by the prompt. For each prompt, the model generates a series of possible continuations, which we term generation beams.
        As a result, in the nested list for each prompt, the lists are grouped by the generation beam, since each prompt produces some set number. In the
        final tier of nesting for each generation, the JSONs are grouped by test case. We note that the evaluation client is agnostic to the list
        structure and instead iterates over each JSON payload for a test cases, converting each JSON to a boolean independently, only maintaining
        the list shape. The JSON for each test case containing the following attributes:

        {
            'code': <code to be evaluated>,
            'input': <test input>,
            'output': <test output>,
            'entry_point': <entry point>,
            'language': <language>,

        }

        The JSON is formatted as [[[request]]] so that the client can batch requests. The outermost list is for the generations of a
        given prompt, the middle list is for the beam generations of a given prompt, and the innermost list is for each test cases.
        Args:
            payload: the materials of the batched HTTPS request to the client organized by prompt, beam generation, and test case.

        Returns:
            Whether the test case passed or failed.
        """
        del payload  # unused
        raise NotImplementedError(f'{type(self).__name__}.invoke is not implemented')

    def close(self):
        """Close the object store."""
        pass



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Log training metadata to [neptune.ai](https://neptune.ai/)."""

__all__ = ['NeptuneLogger']

import os
import pathlib
import warnings
from functools import partial
from importlib.metadata import version
from typing import TYPE_CHECKING, Any, Literal, Optional, Sequence, Union

import numpy as np
import torch
from packaging.version import Version

from composer._version import __version__
from composer.loggers import LoggerDestination
from composer.utils import MissingConditionalImportError, dist

if TYPE_CHECKING:
    from composer import Logger
    from composer.core import State

NEPTUNE_MODE_TYPE = Literal['async', 'sync', 'offline', 'read-only', 'debug']
NEPTUNE_VERSION_WITH_PROGRESS_BAR = Version('1.9.0')


class NeptuneLogger(LoggerDestination):
    """Log to `neptune.ai <https://neptune.ai/>`_.

    For instructions, see the
    `integration guide <https://docs.neptune.ai/integrations/mosaicml_composer/>`_.

    Args:
        project (str, optional): The name of your Neptune project,
            in the form "workspace-name/project-name". If you leave it empty, the
            ``NEPTUNE_PROJECT`` environment variable will be used.
        api_token (str, optional): Your Neptune API token.
            You can leave out this argument if you save your token to the
            ``NEPTUNE_API_TOKEN`` environment variable (recommended).
            You can find your API token in the user menu of the Neptune web app.
        rank_zero_only (bool): Whether to log only on the rank-zero process (default: ``True``).
        upload_checkpoints (bool): Whether the logger should upload checkpoints to Neptune
            (default: ``False``).
        base_namespace (str, optional): The name of the base namespace where the metadata
            is logged (default: "training").
        neptune_kwargs (dict[str, Any], optional): Any additional keyword arguments to the
            ``neptune.init_run()`` function. For options, see the
            `Run API reference <https://docs.neptune.ai/api/neptune/#init_run>`_.
    """
    metric_namespace = 'metrics'
    hyperparam_namespace = 'hyperparameters'
    trace_namespace = 'traces'
    integration_version_key = 'source_code/integrations/neptune-MosaicML'

    def __init__(
        self,
        *,
        project: Optional[str] = None,
        api_token: Optional[str] = None,
        rank_zero_only: bool = True,
        upload_checkpoints: bool = False,
        base_namespace: str = 'training',
        mode: Optional[NEPTUNE_MODE_TYPE] = None,
        **neptune_kwargs,
    ) -> None:
        try:
            from neptune.internal.utils import verify_type
        except ImportError as e:
            raise MissingConditionalImportError(
                extra_deps_group='neptune',
                conda_package='neptune',
                conda_channel='conda-forge',
            ) from e

        verify_type('project', project, (str, type(None)))
        verify_type('api_token', api_token, (str, type(None)))
        verify_type('rank_zero_only', rank_zero_only, bool)
        verify_type('upload_checkpoints', upload_checkpoints, bool)
        verify_type('base_namespace', base_namespace, str)

        if not base_namespace:
            raise ValueError("Argument 'base_namespace' cannot be an empty string.")

        self._project = project
        self._api_token = api_token
        self._rank_zero_only = rank_zero_only

        self._upload_checkpoints = upload_checkpoints

        self._base_namespace = base_namespace
        self._neptune_kwargs = neptune_kwargs

        self._enabled = (not rank_zero_only) or dist.get_global_rank() == 0

        self._mode: Optional[NEPTUNE_MODE_TYPE] = mode if self._enabled else 'debug'

        self._neptune_run = None
        self._base_handler = None

        self._metrics_dict: dict[str, int] = {}  # used to prevent duplicate step logging

        super().__init__()

    @property
    def neptune_run(self):
        """Gets the Neptune run object from a NeptuneLogger instance.

        To log additional metadata to the run, access a path inside the run and assign metadata
        with ``=`` or other `Neptune logging methods <https://docs.neptune.ai/logging/methods/>`_.
        """
        from neptune import Run

        if not self._neptune_run:
            self._neptune_run = Run(
                project=self._project,
                api_token=self._api_token,
                mode=self._mode,
                **self._neptune_kwargs,
            )
        return self._neptune_run

    @property
    def base_handler(self):
        """Gets a handler for the base logging namespace.

        Use the handler to log extra metadata to the run and organize it under the base namespace
        (default: "training"). You can operate on it like a run object: Access a path inside the
        handler and assign metadata to it with ``=`` or other
        `Neptune logging methods <https://docs.neptune.ai/logging/methods/>`_.
        """
        return self.neptune_run[self._base_namespace]

    def init(self, state: 'State', logger: 'Logger') -> None:
        del logger  # unused

        self.base_handler['rank'] = dist.get_global_rank()

        if self._enabled:
            self.neptune_run['sys/name'] = state.run_name
            self.neptune_run[self.integration_version_key] = __version__

    def _sanitize_metrics(self, metrics: dict[str, float], step: Optional[int]) -> dict[str, float]:
        """Sanitize metrics to prevent duplicate step logging.

        Args:
            metrics (dict[str, float]): Metrics to log.
            step (Optional[int]): Step to log metrics at.

        Returns:
            dict[str, float]: Sanitized metrics.
        """
        keys_to_delete: set[str] = set()

        for k in metrics:
            self._process_single_metric(k, step, keys_to_delete)

        return dict(filter(lambda x: x[0] not in keys_to_delete, metrics.items()))

    def _process_single_metric(self, metric_key: str, step: Optional[int], keys_to_delete: set[str]) -> None:
        if metric_key not in self._metrics_dict:
            self._metrics_dict[metric_key] = step if step is not None else 0
        else:
            if step is not None:
                if step <= self._metrics_dict[metric_key]:
                    # we cannot insert metrics earlier than or in place of an existing metric point
                    keys_to_delete.add(metric_key)
                else:
                    self._metrics_dict[metric_key] = step
            else:
                self._metrics_dict[metric_key] += 1

    def log_metrics(self, metrics: dict[str, float], step: Optional[int] = None) -> None:
        if not self._enabled:
            return

        from neptune.utils import stringify_unsupported

        if metrics_to_log := self._sanitize_metrics(metrics, step):
            self.base_handler[NeptuneLogger.metric_namespace].append(stringify_unsupported(metrics_to_log), step=step)

    def log_hyperparameters(self, hyperparameters: dict[str, Any]) -> None:
        if not self._enabled:
            return

        from neptune.utils import stringify_unsupported

        self.base_handler[NeptuneLogger.hyperparam_namespace] = stringify_unsupported(hyperparameters)

    def log_traces(self, traces: dict[str, Any]):
        if not self._enabled:
            return

        from neptune.utils import stringify_unsupported

        self.base_handler[NeptuneLogger.trace_namespace] = stringify_unsupported(traces)

    def can_upload_files(self) -> bool:
        """Whether the logger supports uploading files."""
        return self._enabled and self._upload_checkpoints

    def upload_file(
        self,
        state: 'State',
        remote_file_name: str,
        file_path: pathlib.Path,
        *,
        overwrite: bool = False,
    ):
        if not self.can_upload_files():
            return

        if file_path.is_symlink() or file_path.suffix.lower() == '.symlink':
            return  # skip symlinks

        neptune_path = f'{self._base_namespace}/{remote_file_name}'
        if self.neptune_run.exists(neptune_path) and not overwrite:

            warnings.warn(
                f"The file '{neptune_path}' already exists and overwrite is set to False."
                'No action will be taken.',
            )
            return

        del state  # unused

        from neptune.types import File

        with open(str(file_path), 'rb') as fp:
            self.base_handler[remote_file_name] = File.from_stream(fp, extension=file_path.suffix)

    def download_file(
        self,
        remote_file_name: str,
        destination: str,
        overwrite: bool = False,
        progress_bar: bool = True,
    ):

        if not self._enabled:
            return

        if os.path.exists(
            os.path.join(
                destination,
                remote_file_name,
            ),
        ) and not overwrite:
            warnings.warn(
                f"Destination '{destination}' already exists and overwrite is set to False."
                'No action will be taken.',
            )
            return

        file_path = f'{self._base_namespace}/{remote_file_name}'
        if not self.neptune_run.exists(file_path):
            raise FileNotFoundError(f'File {file_path} not found')

        if _is_progress_bar_enabled():
            self.base_handler[remote_file_name].download(destination=destination, progress_bar=progress_bar)
        else:
            del progress_bar
            self.base_handler[remote_file_name].download(destination=destination)

    def log_images(
        self,
        images: Union[np.ndarray, torch.Tensor, Sequence[Union[np.ndarray, torch.Tensor]]],
        name: str = 'Images',
        channels_last: bool = False,
        step: Optional[int] = None,
        masks: Optional[dict[str, Union[np.ndarray, torch.Tensor, Sequence[Union[np.ndarray, torch.Tensor]]]]] = None,
        mask_class_labels: Optional[dict[int, str]] = None,
        use_table: bool = True,
    ):
        if not self._enabled:
            return

        from neptune.types import File

        if not isinstance(images, Sequence) and images.ndim <= 3:
            images = _validate_image(images, channels_last=channels_last)
            self.base_handler[name].append(File.as_image(images), step=step)

        else:
            images = list(map(partial(_validate_image, channels_last=channels_last), images))
            self.base_handler[name].extend([File.as_image(img) for img in images])

    def post_close(self) -> None:
        if not self._enabled:
            return

        if self._neptune_run:
            self._neptune_run.stop()
            self._neptune_run = None


def _validate_image(img: Union[np.ndarray, torch.Tensor], channels_last: bool) -> np.ndarray:
    img_numpy = img.data.cpu().numpy() if isinstance(img, torch.Tensor) else img

    assert isinstance(img_numpy, np.ndarray)

    # Error out for empty arrays or weird arrays of dimension 0.
    if np.any(np.equal(img_numpy.shape, 0)):
        raise ValueError(f'Got an image (shape {img_numpy.shape}) with at least one dimension being 0! ')

    if not channels_last:
        img_numpy = np.moveaxis(img_numpy, 0, -1)

    return _validate_image_value_range(img_numpy)


def _validate_image_value_range(img: np.ndarray) -> np.ndarray:
    array_min = img.min()
    array_max = img.max()

    if (array_min >= 0 and 1 < array_max <= 255) or (array_min >= 0 and array_max <= 1):
        return img

    from neptune.common.warnings import NeptuneWarning, warn_once

    warn_once(
        'Image value range is not in the expected range of [0.0, 1.0] or [0, 255]. '
        'This might be due to the presence of `transforms.Normalize` in the data pipeline. '
        'Logged images may not display correctly in Neptune.',
        exception=NeptuneWarning,
    )

    return _scale_image_to_0_255(img, array_min, array_max)


def _scale_image_to_0_255(img: np.ndarray, array_min: Union[int, float], array_max: Union[int, float]) -> np.ndarray:
    scaled_image = 255 * (img - array_min) / (array_max - array_min)
    return scaled_image.astype(np.uint8)


def _is_progress_bar_enabled() -> bool:
    return Version(version('neptune')) >= NEPTUNE_VERSION_WITH_PROGRESS_BAR



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

from collections import Counter

from torch.optim.lr_scheduler import (
    CosineAnnealingLR,
    CosineAnnealingWarmRestarts,
    ExponentialLR,
    LRScheduler,
    MultiStepLR,
    StepLR,
)


def scale_pytorch_scheduler(scheduler: LRScheduler, ssr: float):
    """Makes a learning rate schedule take a different number of epochs.

    Training for less time is a strong baseline approach to speeding up
    training, provided that the training still gets through the entire
    learning rate schedule. E.g., training for half as long often yields
    little accuracy degredation, provided that the learning rate schedule
    is rescaled to take half as long as well. In contrast, if the schedule
    is not rescaled, training for half as long would mean simply stopping
    halfway through the training curve, which does reach nearly as
    high an accuracy.

    To see the difference, consider training for half as long using a cosine
    annealing learning rate schedule. If the schedule is not rescaled,
    training ends while the learning rate is still ~0.5. If the schedule is
    rescaled, training ends after passing through the full cosine
    curve, at a learning rate near 0.

    .. doctest::

        >>> from composer.trainer._scale_schedule import scale_scheduler
        >>> from torch.optim.lr_scheduler import CosineAnnealingLR
        >>> scheduler = CosineAnnealingLR(optimizer, T_max=90)
        >>> scheduler = scale_scheduler(scheduler, ssr=0.5)

    Args:
        scheduler: A learning rate schedule object. Must be one of:

            * ``torch.optim.lr_scheduler.CosineAnnealingLR``
            * ``torch.optim.lr_scheduler.CosineAnnealingWarmRestarts``
            * ``torch.optim.lr_scheduler.ExponentialLR``
            * ``torch.optim.lr_scheduler.MultiStepLR``
            * ``torch.optim.lr_scheduler.StepLR``

        ssr: the factor by which to scale the duration of the schedule. E.g., 0.5
            makes the schedule take half as many epochs and 2.0 makes it
            take twice as many epochs.

    Raises:
        ValueError: If ``scheduler`` is not an instance of one of the above types.
    """
    if ssr <= 0:
        raise ValueError('Scale schedule ratio must be a positive value.')
    if ssr == 1.0:
        # If it's 1.0, then scaling is a no-op
        return
    # Applies SSR on a pytorch scheduler in place.
    if isinstance(scheduler, StepLR):
        scheduler.step_size = int(scheduler.step_size * ssr)  # type: ignore  -- unknown attribute
    elif isinstance(scheduler, MultiStepLR):
        milestones = scheduler.milestones  # type: ignore  -- unknown attribute
        milestones = Counter([int(ms * ssr) for ms in milestones])
        scheduler.milestones = milestones  # type: ignore  -- unknown attribute
    elif isinstance(scheduler, CosineAnnealingLR):
        scheduler.T_max = int(scheduler.T_max * ssr)  # type: ignore  -- unknown attribute
    elif isinstance(scheduler, CosineAnnealingWarmRestarts):
        # TODO: account for warmups
        scheduler.T_0 = int(scheduler.T_0 * ssr)  # type: ignore  -- unknown attribute
    elif isinstance(scheduler, ExponentialLR):
        factor = 1 / ssr
        scheduler.gamma = scheduler.gamma**factor  # type: ignore  -- unknown attribute
    else:
        raise ValueError(
            f'Scale schedule being applied to unrecognized Scheduler {scheduler}. '
            'Please implement your scheduler as a function instead.',
        )



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""The HPU device used for training."""

from __future__ import annotations

import logging
from typing import TypeVar

import torch

from composer.devices.device import Device

logger = logging.getLogger(__name__)

__all__ = ['DeviceHPU']

T_nnModule = TypeVar('T_nnModule', bound=torch.nn.Module)


class DeviceHPU(Device):
    """An extension of :class:`~composer.devices.device.Device` for HPUs.

    This class takes no arguments.
    """

    dist_backend = 'hccl'
    name = 'hpu'
    _device = torch.device('hpu')

    def module_to_device(self, module: T_nnModule) -> T_nnModule:
        return module.to(self._device)

    def tensor_to_device(self, tensor: torch.Tensor) -> torch.Tensor:
        return tensor.to(self._device)



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Sequence length warmup progressively increases the sequence length during training of NLP models.

See the :doc:`Method Card </method_cards/seq_length_warmup>` for more details.
"""

from composer.algorithms.seq_length_warmup.seq_length_warmup import SeqLengthWarmup, set_batch_sequence_length

__all__ = ['SeqLengthWarmup', 'set_batch_sequence_length']



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

from composer import (
    algorithms,
    callbacks,
    core,
    devices,
    functional,
    loggers,
    loss,
    metrics,
    models,
    optim,
    profiler,
    trainer,
    utils,
)


# This very simple test is just to use the above imports, which check and make sure we can import all the top-level
# modules from composer. This is mainly useful for checking that we have correctly conditionally imported all optional
# dependencies.
def test_smoketest():
    assert callbacks
    assert algorithms
    assert core
    assert devices
    assert functional
    assert loggers
    assert loss
    assert metrics
    assert models
    assert optim
    assert profiler
    assert trainer
    assert utils



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Log model outputs and expected outputs during ICL evaluation."""

import warnings
from copy import deepcopy
from typing import Any, Sequence, Union

import torch

from composer.core import Callback, State
from composer.loggers import ConsoleLogger, Logger
from composer.utils import VersionedDeprecationWarning, dist


class EvalOutputLogging(Callback):
    """Logs eval outputs for each sample of each ICL evaluation dataset.

    ICL metrics are required to support caching the model's responses including information on whether model was correct.
    Metrics are responsible for returning the results of individual data points in a dictionary of lists.
    The callback will log the metric name, the depadded and detokenized input, any data stored in state.metric_outputs, and
    any keys from the batch passed into `batch_keys_to_log`. It will do so after every eval batch.
    """

    def __init__(self, log_tokens=False, *args, **kwargs):
        warnings.warn(
            VersionedDeprecationWarning(
                '`InContextLearningMetric` and it\'s subclasses have been deprecated and ' +
                'migrated to MosaicML\'s llm-foundry repo under the llmfoundry.eval.datasets.in_context_learning module: '
                + 'https://github.com/mosaicml/llm-foundry/blob/main/scripts/eval/README.md.' +
                'As EvalOutputLogging only works for ICL metrics, it has been deprecated and ' +
                'will be migrated as well.',
                remove_version='0.24.0',
            ),
        )
        super().__init__(self, *args, **kwargs)
        self.log_tokens = log_tokens
        self.columns = None
        self.name = None
        self.rows = []

    def eval_batch_end(self, state: State, logger: Logger) -> None:
        if not isinstance(state.batch, dict):
            warnings.warn(
                f'''EvalOutputLogging only supports batches that are dictionary. \
                Found batch for type {type(state.batch)}. \
                Not logging eval outputs.''',
            )
            return

        assert state.outputs is not None
        assert state.metric_outputs is not None
        logging_dict: dict[str, Union[list[Any], torch.Tensor, Sequence[torch.Tensor]]] = deepcopy(state.metric_outputs)

        # If batch mode is not generate, outputs will be logits
        if state.batch['mode'] == 'generate':
            # Outputs are already detokenized
            logging_dict['outputs'] = state.outputs

        input_ids = state.batch['input_ids']
        logged_input = []
        assert state.dataloader is not None

        # Depad and decode input_ids
        for input_list in input_ids.tolist():
            dataset = state.dataloader.dataset  # pyright: ignore[reportGeneralTypeIssues]
            depadded_input = [tok for tok in input_list if tok != dataset.pad_tok_id]
            logged_input.append(dataset.tokenizer.decode(depadded_input))
        logging_dict['input'] = logged_input

        # Log token indices if toggled
        if self.log_tokens:
            logging_dict['input_tokens'] = input_ids.tolist()
            if not state.batch['mode'] == 'generate':
                if isinstance(state.outputs, torch.Tensor):  # pyright
                    logging_dict['label_tokens'] = state.outputs.tolist()

        # Add run_name as a column
        run_name_list = [state.run_name for _ in range(0, len(logging_dict['input']))]
        logging_dict['run_name'] = run_name_list

        # NOTE: This assumes _any_ tensor logged are tokens to be decoded.
        #       This might not be true if, for example, logits are logged.

        # Detokenize data in rows
        for key, value in logging_dict.items():
            # All types in list are the same
            if isinstance(value[0], torch.Tensor):
                logging_dict[key] = [
                    state.dataloader.dataset.tokenizer.decode(t)  # pyright: ignore[reportGeneralTypeIssues]
                    for t in value
                ]
            elif isinstance(value[0], list):
                if isinstance(value[0][0], torch.Tensor):
                    tokenizer = state.dataloader.dataset.tokenizer  # pyright: ignore[reportGeneralTypeIssues]
                    logging_dict[key] = [[tokenizer.decode(choice) for choice in t] for t in value]

        # Convert logging_dict from kv pairs of column name and column values to a list of rows
        # Example:
        # logging_dict = {"a": ["1a", "2a"], "b": ["1b", "2b"]}
        # will become
        # columns = {"a", "b"}, rows = [["1a", "1b"], ["2a", "2b"]]
        columns = list(logging_dict.keys())
        rows = [list(item) for item in zip(*logging_dict.values())]

        assert state.dataloader_label is not None
        if not self.name:
            # If only running eval, step will be 0
            # If running training, step will be current training step
            step = state.timestamp.batch.value
            self.name = f'{state.dataloader_label}_step_{step}'
            self.columns = columns
        self.rows.extend(rows)

    def eval_end(self, state: State, logger: Logger) -> None:
        list_of_rows = dist.all_gather_object(self.rows)
        rows = [row for rows in list_of_rows for row in rows]
        for dest_logger in logger.destinations:
            if not isinstance(dest_logger, ConsoleLogger):
                dest_logger.log_table(self.columns, rows, name=self.name, step=state.timestamp.batch.value)

        self.rows = []
        self.name = None
        self.columns = None



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Core Label Smoothing classes and functions."""

from __future__ import annotations

from typing import Any, Callable, Optional, Union

import torch

from composer.core import Algorithm, Event, State
from composer.loggers import Logger
from composer.loss.utils import ensure_targets_one_hot

__all__ = ['LabelSmoothing', 'smooth_labels']


def smooth_labels(logits: torch.Tensor, target: torch.Tensor, smoothing: float = 0.1):
    """Shrink targets towards a uniform distribution as in `Szegedy et al <https://arxiv.org/abs/1512.00567>`_.

    The smoothed labels are computed as ``(1 - smoothing) * targets + smoothing * unif``
    where ``unif`` is a vector with elements all equal to ``1 / num_classes``.

    Args:
        logits (torch.Tensor): predicted value for ``target``, or any other tensor
            with the same shape. Shape must be ``(N, num_classes, ...)`` for
            ``N`` examples and ``num_classes`` classes with any number of
            optional extra dimensions.
        target (torch.Tensor): target tensor of either shape ``N`` or
            ``(N, num_classes, ...)``. In the former case, elements of
            ``targets`` must be integer class ids in the range
            ``0..num_classes``. In the latter case, ``targets`` must have the
            same shape as ``logits``.
        smoothing (float, optional): strength of the label smoothing, in
            :math:`[0, 1]`. ``smoothing=0`` means no label smoothing, and
            ``smoothing=1`` means maximal smoothing (targets are ignored).
            Default: ``0.1``.

    Returns:
        torch.Tensor: The smoothed targets.

    Example:
        .. testcode::

            import torch

            num_classes = 10
            targets = torch.randint(num_classes, size=(100,))
            from composer.algorithms.label_smoothing import smooth_labels
            new_targets = smooth_labels(logits=logits,
                                        target=targets,
                                        smoothing=0.1)
    """

    target = ensure_targets_one_hot(logits, target)
    n_classes = logits.shape[1]
    return (target * (1. - smoothing)) + (smoothing / n_classes)


class LabelSmoothing(Algorithm):
    """Shrink targets towards a uniform distribution as in `Szegedy et al <https://arxiv.org/abs/1512.00567>`_.

    The smoothed labels are computed as ``(1 - smoothing) * targets + smoothing * unif``
    where ``unif`` is a vector with elements all equal to ``1 / num_classes``.

    Args:
        smoothing: Strength of the label smoothing, in :math:`[0, 1]`.
            ``smoothing=0`` means no label smoothing, and
            ``smoothing=1`` means maximal smoothing (targets are ignored).
            Default: ``0.1``.
        target_key (str | int | tuple[Callable, Callable] | Any, optional): A key that indexes to the target
            from the batch. Can also be a pair of get and set functions, where the getter
            is assumed to be first in the pair. The default is 1, which corresponds to any sequence, where the second element
            is the target. Default: ``1``.

    Example:
        .. testcode::

            from composer.algorithms import LabelSmoothing
            algorithm = LabelSmoothing(smoothing=0.1)
            trainer = Trainer(
                model=model,
                train_dataloader=train_dataloader,
                eval_dataloader=eval_dataloader,
                max_duration="1ep",
                algorithms=[algorithm],
                optimizers=[optimizer]
            )
    """

    def __init__(
        self,
        smoothing: float = 0.1,
        target_key: Union[str, int, tuple[Callable, Callable], Any] = 1,
    ):
        self.smoothing = smoothing
        self.original_labels = torch.Tensor()
        self.target_key = target_key

    def match(self, event: Event, state: State) -> bool:
        return event in [Event.BEFORE_LOSS, Event.AFTER_LOSS]

    def apply(self, event: Event, state: State, logger: Logger) -> Optional[int]:
        labels = state.batch_get_item(self.target_key)

        if event == Event.BEFORE_LOSS:
            assert isinstance(state.outputs, torch.Tensor), 'Multiple tensors not supported yet'
            assert isinstance(labels, torch.Tensor), 'Multiple tensors not supported yet'

            self.original_labels = labels.clone()
            smoothed_labels = smooth_labels(
                state.outputs,
                labels,
                smoothing=self.smoothing,
            )
            state.batch_set_item(self.target_key, smoothed_labels)
        elif event == Event.AFTER_LOSS:
            # restore the target to the non-smoothed version
            state.batch_set_item(self.target_key, self.original_labels)



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Monitor activation values during training."""

import warnings
from functools import partial
from typing import Any, Optional, Sequence, Union

import torch

from composer.core import Callback, State, Time, TimeUnit
from composer.loggers import Logger
from composer.loggers.wandb_logger import WandBLogger

__all__ = ['ActivationMonitor']


class ActivationMonitor(Callback):
    """Logs stats of activation inputs and outputs.

    This callback triggers at a user defined interval, and logs some simple statistics of the inputs, outputs for every
    torch module. This is done by attaching a forward hook to the module. Additionally, when after we finish logging
    we detach the forwards hook.

    Example:
        .. doctest::

            >>> from composer import Trainer
            >>> from composer.callbacks import ActivationMonitor
            >>> # constructing trainer object with this callback
            >>> trainer = Trainer(
            ...     model=model,
            ...     train_dataloader=train_dataloader,
            ...     eval_dataloader=eval_dataloader,
            ...     optimizers=optimizer,
            ...     max_duration="1ep",
            ...     callbacks=[ActivationMonitor()],
            ... )

    The metrics are logged by the :class:`.Logger` to the following keys described below. Over an input of shape
    (batch, hid_dim), (batch, seq_len, hid_dim), etc. we compute statistics across `hid_dim` then take the average
    of these statistics. For convenience we have included example metrics logged:

        +-------------------------------------------------------+-----------------------------------------------------+
        | Key                                                   | Logged data                                         |
        +=======================================================+=====================================================+
        |                                                       | The average max value of the `hid_dim` of the       |
        | ``activations/max/MODULE_NAME/input_{n}``             | nth input activations into the current module.      |
        |                                                       |                                                     |
        +-------------------------------------------------------+-----------------------------------------------------+
        |                                                       | The average value of the `hid_dim` input            |
        | ``activations/average/MODULE_NAME/input_{n}``         | activations into the current module.                |
        |                                                       |                                                     |
        +-------------------------------------------------------+-----------------------------------------------------+
        |                                                       | The average L2 Norm of the `hid_dim` of the         |
        | ``activations/l2_norm/MODULE_NAME/input_{n}``         | nth input activations into the current module.      |
        |                                                       |                                                     |
        +-------------------------------------------------------+-----------------------------------------------------+
        |                                                       | The average kurtosis of the `hid_dim` of the nth    |
        | ``activations/kurtosis/MODULE_NAME/input_{n}``        | input activations into the current module.          |
        |                                                       |                                                     |
        +-------------------------------------------------------+-----------------------------------------------------+
        |                                                       | The average max value of the `hid_dim` of the       |
        | ``activations/max/MODULE_NAME/output_{n}``            | nth ouput activations of the current module.        |
        |                                                       |                                                     |
        +-------------------------------------------------------+-----------------------------------------------------+
        |                                                       | The average value of the `hid_dim` of the output    |
        | ``activations/average/MODULE_NAME/output_{n}``        | activations of the current module.                  |
        |                                                       |                                                     |
        +-------------------------------------------------------+-----------------------------------------------------+
        |                                                       | The average L2 Norm of the values of the `hid_dim`  |
        | ``activations/l2_norm/MODULE_NAME/input_{n}``         | activations of the current module.                  |
        |                                                       |                                                     |
        +-------------------------------------------------------+-----------------------------------------------------+
        |                                                       | The average kurtosis of the `hid_dim` of the nth    |
        | ``activations/kurtosis/MODULE_NAME/input_{n}``        | output activations of the current module.           |
        |                                                       |                                                     |
        +-------------------------------------------------------+-----------------------------------------------------+

    Args:
        interval (Union[int, str, Time], optional): Time string specifying how often to attach the logger and log the activations.
            For example, ``interval='5ba'`` means every 5 batches we log the activations. Default: '25ba'.
        ignore_module_types (Optional[list[str]], optional): A list of strings representing the class attributes we should ignore.
            For example passing in the list ['dropout', 'ln'] will cause the class attributes that contain
            'dropout' or 'ln' to not be logged. Default: 'None'.
        only_log_wandb (bool, optional): A bool that determines if we should only log to Weights and Biases. This is recommended
            in partcular for larger models as this callback logs a lot. Default: 'True'.
    """

    def __init__(
        self,
        interval: Union[int, str, Time] = '25ba',
        ignore_module_types: Optional[list[str]] = None,
        only_log_wandb: bool = True,
    ):
        self.ignore_module_types = ignore_module_types
        self.only_log_wandb = only_log_wandb

        self.handles = []

        # Check that the interval timestring is parsable and convert into time object
        self.interval = Time.from_input(interval, TimeUnit.BATCH)

        if self.interval.unit == TimeUnit.BATCH and self.interval < Time.from_timestring('10ba'):
            warnings.warn(
                f'Currently the ActivationMonitor`s interval is set to {self.interval} '
                f'which is below our recommended value of 10ba. We recommend you raise '
                f'the interval to at least 10ba, as the activation monitor adds extra overhead '
                f'and decreases throughput.',
            )

        # Verify that the interval has supported units
        if self.interval.unit not in [TimeUnit.BATCH, TimeUnit.EPOCH]:
            raise ValueError(
                f'Invalid time unit for parameter interval: '
                f'{self.interval.unit}',
            )

        self.last_train_time_value_logged = -1
        self.module_names = {}

    def before_forward(self, state: State, logger: Logger):
        current_time_value = state.timestamp.get(self.interval.unit).value

        if current_time_value % self.interval.value == 0 and current_time_value != self.last_train_time_value_logged:
            if not self.module_names:
                self.create_module_names(state.model)

            self.attach_forward_hooks(state, logger)

    def after_forward(self, state: State, logger: Logger):
        current_time_value = state.timestamp.get(self.interval.unit).value

        if current_time_value % self.interval.value == 0 and current_time_value != self.last_train_time_value_logged:
            self.last_train_time_value_logged = current_time_value
            self.remove_forward_hooks()

    def attach_forward_hooks(self, state: State, logger: Logger):
        step = state.timestamp.batch.value
        self.register_forward_hook(state.model, logger, step)

    def remove_forward_hooks(self):
        for handle in self.handles:
            handle.remove()
        # Resetting handles we track
        self.handles = []

    def register_forward_hook(self, model: torch.nn.Module, logger: Logger, step: Optional[int]):
        model.apply(partial(self._register_forward_hook, logger, step))

    def _register_forward_hook(self, logger: Logger, step: Optional[int], module: torch.nn.Module):
        self.handles.append(module.register_forward_hook(partial(self.forward_hook, logger, step)))

    def forward_hook(
        self,
        logger: Logger,
        step: Optional[int],
        module: torch.nn.Module,
        input: Optional[Sequence],
        output: Optional[Sequence],
    ):
        module_name = self.module_names[module]

        if self.ignore_module_types is not None:
            for ignore_module_type in self.ignore_module_types:
                if ignore_module_type in module_name:
                    return

        metrics = {}
        if input is not None:
            for i, val in enumerate(input):
                if val is None or isinstance(val, dict):
                    continue
                if isinstance(val, str) and isinstance(input, dict):
                    self.recursively_add_metrics(metrics, module_name, f'_input.{i}', output[val])  # type: ignore
                else:
                    self.recursively_add_metrics(metrics, module_name, f'_input.{i}', val)

        if output is not None:
            for i, val in enumerate(output):
                if val is None or isinstance(val, dict):
                    continue
                if isinstance(val, str) and isinstance(output, dict):
                    self.recursively_add_metrics(metrics, module_name, f'_output.{i}', output[val])  # type: ignore
                else:
                    self.recursively_add_metrics(metrics, module_name, f'_output.{i}', val)

        if self.only_log_wandb:
            wandb_loggers = [ld for ld in logger.destinations if isinstance(ld, WandBLogger)]
            if len(wandb_loggers):
                for wandb_logger in wandb_loggers:
                    wandb_logger.log_metrics(metrics, step)
            else:
                # In the case there were no WandB loggers, just default to
                # the standard logger and let it take care of it
                logger.log_metrics(metrics)
        else:
            logger.log_metrics(metrics)

    def recursively_add_metrics(self, metrics: dict, name: str, suffix: str, values: Any):
        # Becuase of the recursive diving, we need this call to prevent infinite recursion.
        if isinstance(values, str):
            return
        # Keep recursively diving if the value is a sequence
        if isinstance(values, Sequence):
            for i, value in enumerate(values):
                self.recursively_add_metrics(metrics, f'{name}_{i}', suffix, value)
            return
        else:
            self.add_metrics(metrics, name, suffix, values)

    def add_metrics(self, metrics: dict, name: str, suffix: str, value: torch.Tensor):
        # We shouldn't log booleans
        if value.dtype == torch.bool:
            return
        if value.is_floating_point() or value.is_complex():
            metrics[f'activations/l2_norm/{name}{suffix}'] = torch.linalg.vector_norm(value, dim=-1).mean().item()
            metrics[f'activations/average/{name}{suffix}'] = value.mean().item()
            metrics[f'activations/kurtosis/{name}{suffix}'] = compute_kurtosis(value).item()

            # Because we call max with `dim=-1` we need to call .values to get the actual values
            metrics[f'activations/max/{name}{suffix}'] = value.max(dim=-1).values.mean().item()

    def create_module_names(self, model: torch.nn.Module):
        self.module_names = {m: name for name, m in model.named_modules()}


def compute_kurtosis(value: torch.Tensor):
    # Computes the kurtosis over the last dimension
    mean = torch.mean(value, dim=-1).unsqueeze(-1)
    diffs = value - mean
    m_4 = torch.mean(torch.pow(diffs, 4), dim=-1)
    var = torch.mean(torch.pow(diffs, 2), dim=-1)
    return (m_4 / (var**2)).mean()



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""A wrapper for a dataloader to include metrics that apply to a specific dataset."""

from __future__ import annotations

import textwrap
import warnings
from typing import Any, Callable, Iterable, Optional, Union

from composer.core.data_spec import DataSpec, ensure_data_spec
from composer.core.event import Event
from composer.core.state import State
from composer.core.time import Time
from composer.utils import create_interval_scheduler

__all__ = ['Evaluator', 'ensure_evaluator']


class Evaluator:
    """A wrapper for a dataloader to include metrics that apply to a specific dataset.

    For example, :class:`.CrossEntropyLoss` metric for NLP models.

    .. doctest::

       >>> eval_evaluator = Evaluator(
       ...     label='myEvaluator',
       ...     dataloader=eval_dataloader,
       ...     metric_names=['MulticlassAccuracy']
       ... )
       >>> trainer = Trainer(
       ...     model=model,
       ...     train_dataloader=train_dataloader,
       ...     eval_dataloader=eval_evaluator,
       ...     optimizers=optimizer,
       ...     max_duration='1ep',
       ... )

    Args:
        label (str): Name of the Evaluator.
        dataloader (DataSpec | Iterable | dict[str, Any]): Iterable that yields batches, a :class:`.DataSpec`
            for evaluation, or a dict of :class:`.DataSpec` kwargs.
        metric_names: The list of metric names to compute.
            Each value in this list can be a regex string (e.g. "MulticlassAccuracy", "f1" for "BinaryF1Score",
            "Top-." for "Top-1", "Top-2", etc). Each regex string will be matched against the keys of the dictionary returned
            by ``model.get_metrics()``. All matching metrics will be evaluated.

            By default, if left blank, then all metrics returned by ``model.get_metrics()`` will be used.
        subset_num_batches (int, optional): The maximum number of batches to use for each evaluation. Defaults to ``None``,
            which means that the ``eval_subset_num_batches`` parameter from the :class:`.Trainer` will be used.
            Set to ``-1`` to evaluate the entire ``dataloader``.
        eval_interval (Time | int | str | (State, Event) -> bool, optional): An integer,
            which will be interpreted to be epochs, a str (e.g. ``1ep``, or ``10ba``), a :class:`.Time` object, or a callable.
            Defaults to ``None``, which means that the ``eval_interval`` parameter from the :class:`.Trainer` will be used.

            If an integer (in epochs), :class:`.Time` string, or :class:`.Time` instance, the evaluator will be run
            with this frequency. :class:`.Time` strings or :class:`.Time` instances must have units of
            :attr:`.TimeUnit.BATCH` or :attr:`.TimeUnit.EPOCH`.

            Set to ``0`` to disable evaluation.

            If a callable, it should take two arguments (:class:`.State`, :class:`.Event`) and return a bool
            representing whether the evaluator should be invoked. The event will be either :attr:`.Event.BATCH_END`
            or :attr:`.Event.EPOCH_END`.

            When specifying ``eval_interval``, the evaluator(s) are also run at the ``Event.FIT_END`` if it doesn't
            evenly divide the training duration.
        device_eval_microbatch_size (int, optional): The number of samples to use for each microbatch when evaluating.
            If set to ``auto``, dynamically decreases device_eval_microbatch_size if microbatch is too large for GPU.
            If None, sets `device_eval_microbatch_size` to per rank batch size. (default: ``None``)
    """

    def __init__(
        self,
        *,
        label: str,
        dataloader: Union[DataSpec, Iterable, dict[str, Any]],
        metric_names: Optional[list[str]] = None,
        subset_num_batches: Optional[int] = None,
        eval_interval: Optional[Union[int, str, Time, Callable[[State, Event], bool]]] = None,
        device_eval_microbatch_size: Optional[Union[int, str]] = None,
    ):
        self.label = label
        self.dataloader = ensure_data_spec(dataloader)

        if metric_names is not None:
            if not isinstance(metric_names, list):
                raise ValueError(f'``metric_names`` should be a list of strings, not a {type(metric_names)}')
        self.metric_names = metric_names

        self.subset_num_batches = subset_num_batches
        self._eval_interval = None
        self.eval_interval = eval_interval
        self.auto_microbatching = _is_auto_microbatching(device_eval_microbatch_size)
        if self.auto_microbatching and hasattr(self.dataloader, 'seq_parallel_world_size'):
            raise ValueError('`device_eval_microbatch_size="auto"` is not compatible with sequence parallelism.')
        self.device_eval_microbatch_size = _get_initial_device_eval_microbatch_size(
            device_eval_microbatch_size,
            self.auto_microbatching,
            self.dataloader.dataloader,
        )

    @property
    def eval_interval(self):
        return self._eval_interval

    @eval_interval.setter
    def eval_interval(self, eval_interval: Optional[Union[int, str, Time, Callable[[State, Event], bool]]]):
        if eval_interval is None:
            self._eval_interval = None
        elif not callable(eval_interval):
            self._eval_interval = create_interval_scheduler(
                eval_interval,
                checkpoint_events=False,
                final_events={Event.FIT_END},
            )
        else:
            self._eval_interval = eval_interval


def ensure_evaluator(evaluator: Union[Evaluator, DataSpec, Iterable, dict[str, Any]], default_metric_names: list[str]):
    """Ensure that ``evaluator`` is an :class:`.Evaluator`.

    Args:
        evaluator (Evaluator | DataSpec | Iterable | dict[str, Any]): A dataloader,
            :class:`.DataSpec` instance, dictionary of :class:`.DataSpec` kwargs, or existing evaluator.
        default_metric_names (list[str]): The names of the metrics for the ``evaluator``,
            if a dataloader was specified.

    Returns:
        Evaluator: An evaluator.
    """
    if isinstance(evaluator, Evaluator):
        return evaluator
    else:
        return Evaluator(
            label='eval',
            dataloader=evaluator,
            metric_names=default_metric_names,
        )


def _is_auto_microbatching(device_eval_microbatch_size: Optional[Union[int, str]]):
    if device_eval_microbatch_size == 'auto':
        warnings.warn((
            "Setting `device_eval_microbatch_size='auto'` is an experimental feature which may cause "
            'uncaught Cuda Out of Memory errors. In this case, please manually '
            'set device_eval_microbatch_size explicitly to an integer instead.'
        ))
        return True
    else:
        return False


def _get_initial_device_eval_microbatch_size(
    device_eval_microbatch_size: Optional[Union[int, str]],
    auto_microbatching: bool,
    dataloader: Iterable,
) -> int:
    """Sets initial value of device_eval_microbatch_size.

    If auto_microbatching, sets initial `device_eval_microbatch_size` to per rank batch size.
    """
    if auto_microbatching or device_eval_microbatch_size is None:
        try:
            batch_size = getattr(dataloader, 'batch_size')
        except AttributeError as e:
            if auto_microbatching:
                raise AttributeError(
                    "`device_eval_microbatch_size='auto'` requires the `dataloader` to have a `batch_size` attribute.",
                ) from e
            else:
                raise AttributeError(
                    textwrap.dedent(
                        '`device_eval_microbatch_size` is not set and `dataloader` does not have a `batch_size` attribute. '
                        'Please either set `device_eval_microbatch_size` or `dataloader.batch_size`.',
                    ),
                ) from e
        return batch_size
    elif isinstance(device_eval_microbatch_size, (int, float)):
        return device_eval_microbatch_size
    else:
        raise ValueError("device_eval_microbatch_size must be an int or ``'auto'``")



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""System metrics monitor callback."""

from __future__ import annotations

import logging
import os

import psutil
import torch

from composer.core import Callback, Event, State
from composer.loggers import Logger
from composer.utils import dist
from composer.utils.import_helpers import MissingConditionalImportError

log = logging.getLogger(__name__)

__all__ = ['SystemMetricsMonitor']

_GPU_METRICS = [
    'gpu_percentage',
    'memory_percentage',
    'gpu_temperature_C',
    'gpu_power_usage_W',
]


class SystemMetricsMonitor(Callback):
    """Logs GPU/CPU metrics.

    GPU Metrics:
        gpu_percentage: Occupancy rate, percent of time over sampling period during which one or more kernels was executing on the GPU.
        memory_percentage: Percent of time over sampling period during which global memory was being read or written.
        gpu_temperature_C: Temperature of device, in Celcius.
        gpu_power_usage_W: Power usage of device, in Watts.

    By default, only the maximum and minimum values for these metrics, alongside their respective ranks in the key names,
    are logged on the :attr:`.Event.BATCH_START`, :attr:`.Event.EVAL_BATCH_START`, :attr:`.Event.PREDICT_BATCH_START`
    events for every batch. If log_all_data is set to True, all values for these metrics across all ranks are logged on the
    above events for every batch.

    Example:
    .. doctest::

        >>> from composer import Trainer
        >>> from composer.callbacks import SystemMetricsMonitor
        >>> # constructing trainer object with this callback
        >>> trainer = Trainer(
        ...    model=model,
        ...    train_dataloader=train_dataloader,
        ...    eval_dataloader=eval_dataloader,
        ...    optimizers=optimizer,
        ...    max_duration='1ep',
        ...    callbacks=[SystemMetricsMonitor()],
        ... )

    Args:
        log_all_data (bool, optional): True if user wants to log data for all ranks, not just the min/max.
            Defaults to False.
    """

    def __init__(self, log_all_data: bool = False) -> None:
        super().__init__()
        self.gpu_available = torch.cuda.is_available()
        self.log_all_data = log_all_data
        if self.gpu_available:
            try:
                import pynvml
            except ImportError as e:
                raise MissingConditionalImportError(
                    extra_deps_group='pynvml',
                    conda_package='pynvml',
                    conda_channel='conda-forge',
                ) from e
            pynvml.nvmlInit()

    def run_event(self, event: Event, state: State, logger: Logger):
        # only run on the following events
        if event in [
            Event.BATCH_START,
            Event.EVAL_BATCH_START,
            Event.PREDICT_BATCH_START,
        ]:
            local_node_system_metrics = self.compute_system_metrics()
            all_system_metrics = dist.all_gather_object(local_node_system_metrics)
            system_metrics = {}

            if self.log_all_data:
                for rank, metrics in enumerate(all_system_metrics):
                    for key, value in metrics.items():
                        if key in _GPU_METRICS:
                            system_metrics[f'{key}_rank_{rank}'] = value
                        else:
                            system_metrics[key] = value

            else:
                system_metrics = self.compute_gpu_min_max_metrics(all_system_metrics, state)
                for rank, metrics in enumerate(all_system_metrics):
                    for key, value in metrics.items():
                        if key not in _GPU_METRICS:
                            system_metrics[key] = value

            logger.log_metrics(system_metrics)

    def compute_system_metrics(self):
        system_metrics = {}

        # Get metrics for this device if available
        if self.gpu_available:
            import pynvml
            local_rank = dist.get_local_rank()
            handle = pynvml.nvmlDeviceGetHandleByIndex(local_rank)
            device_utilization = pynvml.nvmlDeviceGetUtilizationRates(handle)
            system_metrics['gpu_percentage'] = device_utilization.gpu
            system_metrics['memory_percentage'] = device_utilization.memory
            temperature = pynvml.nvmlDeviceGetTemperature(handle, pynvml.NVML_TEMPERATURE_GPU)
            system_metrics['gpu_temperature_C'] = temperature
            power = pynvml.nvmlDeviceGetPowerUsage(handle) / 1000.0  # convert from mW to W
            system_metrics['gpu_power_usage_W'] = power

        # Get metrics for the system
        cpu_percent = psutil.cpu_percent()
        system_metrics[f'cpu_percentage'] = cpu_percent
        system_memory = psutil.virtual_memory()._asdict()
        for k, v in system_memory.items():
            system_metrics[f'cpu_memory_{k}'] = v
        disk_usage = psutil.disk_usage(os.sep)._asdict()
        for k, v in disk_usage.items():
            system_metrics[f'disk_memory_{k}'] = v
        network_usage = psutil.net_io_counters()._asdict()
        for k, v in network_usage.items():
            system_metrics[f'network_{k}'] = v
        return system_metrics

    def compute_gpu_min_max_metrics(self, all_metrics, state):
        min_max_metrics = {}

        if self.gpu_available:
            for key in _GPU_METRICS:
                values = torch.tensor([metrics_for_cur_rank[key] for metrics_for_cur_rank in all_metrics])
                values = state.device.tensor_to_device(values)
                min_rank = int(torch.argmin(values).item())
                max_rank = int(torch.argmax(values).item())
                min_max_metrics[f'min_{key}_rank_{min_rank}'] = values[min_rank].item()
                min_max_metrics[f'max_{key}_rank_{max_rank}'] = values[max_rank].item()

        return min_max_metrics
